import typing

from bloqade.geometry.dialects import grid
from kirin.lowering import wraps

from bloqade.shuttle.dialects.measure.types import MeasurementArray

from .stmts import Measure

NumX = typing.TypeVar("NumX", bound=int)
NumY = typing.TypeVar("NumY", bound=int)


@wraps(Measure)
def measure(
    regions: tuple[grid.Grid[NumX, NumY], ...],
) -> tuple[MeasurementArray[NumX, NumY], ...]:
    """
    Measure the given regions and return the results.

    Args:
        regions: A tuple of regions to measure.

    Returns:
        A tuple of measurement results.
    """
    ...



from dataclasses import dataclass

from kirin import ir
from kirin.passes import Pass
from kirin.rewrite import Walk
from kirin.rewrite.abc import RewriteResult, RewriteRule

from bloqade.shuttle import spec
from bloqade.shuttle.analysis.zone import Zone, ZoneAnalysis


@dataclass
class ZoneHintRewrite(RewriteRule):
    analysis_results: dict[ir.SSAValue, Zone]

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        has_done_something = False
        for result in node.results:
            if has_done_something := result in self.analysis_results:
                result.hints["zone.analysis"] = self.analysis_results[result]

        return RewriteResult(has_done_something=has_done_something)


@dataclass
class HintZone(Pass):
    """
    This pass adds zone hints to SSA values based on the analysis results.
    """

    arch_spec: spec.ArchSpec

    def unsafe_run(self, mt: ir.Method) -> RewriteResult:
        analysis_frame, _ = ZoneAnalysis(
            mt.dialects, arch_spec=self.arch_spec
        ).run_analysis(mt)
        return Walk(ZoneHintRewrite(analysis_frame.entries)).rewrite(mt.code)



from . import runtime as runtime
from ._dialect import dialect as dialect
from .concrete import PathInterpreter as PathInterpreter
from .constprop import ConstProp as ConstProp
from .spec_interp import SpecPathInterpreter as SpecPathInterpreter
from .stmts import (
    Auto as Auto,
    Gen as Gen,
    Parallel as Parallel,
    Play as Play,
)
from .types import Path as Path, PathType as PathType



from typing import Generic, TypeVar

from kirin import types


# TODO: replace this with the squin dialect's MeasurementResultType
class MeasurementResult:
    pass


NumRows = TypeVar("NumRows")
NumCols = TypeVar("NumCols")


class MeasurementArray(Generic[NumRows, NumCols]):

    def __getitem__(self, indices: tuple[int, int]) -> MeasurementResult:
        """
        Get a measurement result from the array using the given indices.
        """
        raise NotImplementedError(
            "This Class is a placeholder and should be replaced with the actual implementation."
        )


MeasurementResultType = types.PyClass(MeasurementResult)
MeasurementArrayType = types.Generic(
    MeasurementArray, types.TypeVar("NumRows"), types.TypeVar("NumCols")
)



from kirin import ir
from kirin.dialects import func, py
from kirin.rewrite import abc

from bloqade.shuttle.dialects import path, schedule


class RewriteDeviceCall(abc.RewriteRule):
    def rewrite_Statement(self, node: ir.Statement) -> abc.RewriteResult:
        if not isinstance(node, func.Call) or not node.callee.type.is_subseteq(
            schedule.DeviceFunctionType
        ):
            return abc.RewriteResult()

        (
            path_gen := path.Gen(node.callee, node.inputs, kwargs=node.kwargs)
        ).insert_before(node)

        if not isinstance(node.parent_stmt, (schedule.Auto, schedule.Parallel)):
            node.replace_by(path.Play(path_gen.result))
        else:
            node.delete()

        return abc.RewriteResult(has_done_something=True)


class RewriteAutoInvoke(abc.RewriteRule):
    def rewrite_Statement(self, node: ir.Statement) -> abc.RewriteResult:
        if not isinstance(node.parent_stmt, schedule.Auto):
            return abc.RewriteResult()

        if isinstance(node, func.Invoke):
            (callee_stmt := py.Constant(node.callee)).insert_before(node)
            callee_ssa = callee_stmt.result
        elif isinstance(node, func.Call):
            callee_ssa = node.callee
        else:
            return abc.RewriteResult()

        (tweezer_task := schedule.NewTweezerTask(move_fn=callee_ssa)).insert_before(
            node
        )
        (path.Gen(tweezer_task.result, node.inputs, kwargs=node.kwargs)).insert_before(
            node
        )

        node.delete()

        return abc.RewriteResult(has_done_something=True)


class RewriteScheduleRegion(abc.RewriteRule):
    CLASSES = {
        schedule.Auto: path.Auto,
        schedule.Parallel: path.Parallel,
    }

    def rewrite_Statement(self, node: ir.Statement) -> abc.RewriteResult:
        if not isinstance(node, schedule.ExecutableRegion):
            return abc.RewriteResult()

        new_node_type = self.CLASSES[type(node)]
        assert isinstance(node, schedule.ExecutableRegion)
        paths = []

        for stmt in node.body.walk():
            stmt.detach()
            stmt.insert_before(node)
            stmt_result = stmt.expect_one_result()
            if len(stmt_result.uses) == 0:
                # this is required for nested regions where a previously
                # lifted region might have used this current SSA value.
                paths.append(stmt_result)

        (parallel := new_node_type(tuple(paths))).insert_before(node)

        if not isinstance(node.parent_stmt, schedule.ExecutableRegion):
            # If the parent is not a parallel or auto region, we need to wrap it in a Play statement
            # to ensure it can be executed.
            node.replace_by(stmt=path.Play(parallel.result))
        else:
            # If the parent is already a parallel or auto region, we can just replace the node.
            node.delete()

        return abc.RewriteResult(has_done_something=True)


class Canonicalize(abc.RewriteRule):
    """Flatten nested parallel regions into a single parallel region."""

    CLASSES = (schedule.Auto, schedule.Parallel)

    def rewrite_Statement(self, node: ir.Statement) -> abc.RewriteResult:
        if (node_type := type(node)) not in self.CLASSES or not isinstance(
            node.parent_stmt, node_type
        ):
            return abc.RewriteResult()

        assert isinstance(node, self.CLASSES)

        detached_stmts = [
            stmt for stmt in node.body.walk() if not isinstance(stmt, node_type)
        ]
        has_done_something = False
        for stmt in detached_stmts:
            stmt.detach()
            stmt.insert_before(node)
            has_done_something = True

        if len(node.body.blocks[0].stmts) == 0:
            # If the body is empty, we can remove the node
            has_done_something = True
            node.delete()

        return abc.RewriteResult(has_done_something=has_done_something)



from ._dialect import dialect as dialect
from ._interface import (
    measure as measure,
    move as move,
    move_next_to as move_next_to,
    new as new,
    reset_position as reset_position,
)
from .stmts import (
    Measure as Measure,
    Move as Move,
    MoveNextTo as MoveNextTo,
    New as New,
    ResetPosition as ResetPosition,
)
from .types import Atom as Atom, AtomType as AtomType



from dataclasses import dataclass, field
from itertools import product

import numpy as np
from bloqade.geometry.dialects.grid.types import Grid
from matplotlib import pyplot as plt
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.widgets import Button

from bloqade.shuttle.codegen import taskgen
from bloqade.shuttle.dialects import path

from .interface import RendererInterface


def default_ax() -> Axes:
    """Create a default matplotlib Axes object."""
    _, ax = plt.subplots(1, 1)
    ax.set_aspect("equal", adjustable="box")
    ax.set_xlabel("X (um)")
    ax.set_ylabel("Y (um)")
    return ax


@dataclass
class GateDisplayOptions:
    rydberg_color: str = field(default="red")
    r_color: str = field(default="blue")
    rz_color: str = field(default="green")
    local_spot_size: float = field(default=6.0)
    alpha: float = field(default=0.3)


@dataclass
class MatplotlibRenderer(RendererInterface):
    ax: Axes = field(default_factory=default_ax, repr=False)

    gate_display_options: GateDisplayOptions = field(default_factory=GateDisplayOptions)
    arrow_rescale: float = field(default=1.0, kw_only=True)

    active_x_tones: set[int] = field(default_factory=set, repr=False, init=False)
    active_y_tones: set[int] = field(default_factory=set, repr=False, init=False)
    curr_path_lines: list[Artist] = field(default_factory=list, repr=False, init=False)

    xmin: float = field(init=False)
    xmax: float = field(init=False)
    ymin: float = field(init=False)
    ymax: float = field(init=False)
    sleep_time: float = field(default=0.1, kw_only=True)
    sleeping: bool = field(default=True, init=False)

    def __post_init__(self) -> None:
        fig = plt.gcf()
        fig.subplots_adjust(bottom=0.2)
        continue_ax = fig.add_axes((0.01, 0.01, 0.1, 0.075))
        exit_ax = fig.add_axes((0.21, 0.01, 0.1, 0.075))

        self.continue_button = Button(continue_ax, "Continue")
        self.exit_button = Button(exit_ax, "Exit")
        self.exit_button.on_clicked(lambda event: exit())

    @property
    def fov_size(self) -> float:
        return np.sqrt((self.xmax - self.xmin) ** 2 + (self.ymax - self.ymin) ** 2)

    @property
    def arrow_scale(self) -> float:
        return self.fov_size / 400.0 * self.arrow_rescale

    def update_x_bounds(self, y: float) -> None:
        xmin = min(curr_xmin := getattr(self, "xmin", float("inf")), y - 3)
        xmax = max(curr_xmax := getattr(self, "xmax", float("-inf")), y + 3)

        if xmin < curr_xmin or xmax > curr_xmax:
            self.xmin = xmin
            self.xmax = xmax
            self.ax.set_xlim(self.xmin, self.xmax)

    def update_y_bounds(self, y: float) -> None:
        ymin = min(curr_ymin := getattr(self, "ymin", float("inf")), y - 3)
        ymax = max(curr_ymax := getattr(self, "ymax", float("-inf")), y + 3)

        if ymin < curr_ymin or ymax > curr_ymax:
            self.ymin = ymin
            self.ymax = ymax
            self.ax.set_ylim(self.ymin, self.ymax)

    def render_traps(self, traps: Grid, zone_id: str) -> None:
        x, y = np.meshgrid(traps.x_positions, traps.y_positions)

        self.ax.plot(x.flatten(), y.flatten(), marker="o", markersize=3, linestyle="")
        self.update_x_bounds(traps.x_positions[0])
        self.update_x_bounds(traps.x_positions[-1])
        self.update_y_bounds(traps.y_positions[0])
        self.update_y_bounds(traps.y_positions[-1])

    def top_hat_cz(
        self, location: Grid, upper_buffer: float, lower_buffer: float
    ) -> None:
        ymin, ymax = location.y_bounds()

        assert ymin is not None and ymax is not None, "Y bounds must be defined"
        ymin = ymin - lower_buffer
        ymax = ymax + upper_buffer
        ymin_keepout = ymin - 3
        ymax_keepout = ymax + 3
        x = [self.xmin - 10, self.xmax + 10]
        inner = self.ax.fill_between(
            x,
            ymin,
            ymax,
            color=self.gate_display_options.rydberg_color,
            alpha=self.gate_display_options.alpha,
        )
        outer = self.ax.fill_between(
            x,
            ymin_keepout,
            ymax_keepout,
            color=self.gate_display_options.rydberg_color,
            alpha=self.gate_display_options.alpha / 2,
        )

        self.show()
        inner.remove()
        outer.remove()

    def local_r(self, location: Grid) -> None:
        x, y = np.meshgrid(location.x_positions, location.y_positions)

        (points,) = self.ax.plot(
            x.flatten(),
            y.flatten(),
            marker="o",
            markersize=self.gate_display_options.local_spot_size,
            linestyle="",
            color=self.gate_display_options.r_color,
            alpha=self.gate_display_options.alpha,
        )

        self.show()
        points.remove()

    def local_rz(self, location: Grid) -> None:
        x, y = np.meshgrid(location.x_positions, location.y_positions)

        (points,) = self.ax.plot(
            x.flatten(),
            y.flatten(),
            marker="o",
            markersize=self.gate_display_options.local_spot_size,
            linestyle="",
            color=self.gate_display_options.rz_color,
            alpha=self.gate_display_options.alpha,
        )

        self.show()
        points.remove()

    def global_r(self) -> None:
        x = [self.xmin - 10, self.xmax + 10]
        y = [self.ymin - 10, self.ymax + 10]

        line = self.ax.fill_between(
            x,
            y[0],
            y[1],
            color=self.gate_display_options.r_color,
            alpha=self.gate_display_options.alpha,
        )
        self.show()
        line.remove()

    def global_rz(self) -> None:
        x = [self.xmin - 10, self.xmax + 10]
        y = [self.ymin - 10, self.ymax + 10]

        line = self.ax.fill_between(
            x,
            y[0],
            y[1],
            color=self.gate_display_options.rz_color,
            alpha=self.gate_display_options.alpha / 2,
        )

        self.show()
        line.remove()

    def render_path(self, pth: path.Path) -> None:
        all_waypoints = [
            way_point
            for path_action in pth.path
            if isinstance(path_action, taskgen.WayPointsAction)
            for way_point in path_action.way_points
        ]

        num_unique_waypoints = len(set(all_waypoints))
        if num_unique_waypoints < 2:
            return

        num_arrows = num_unique_waypoints - 1
        first_waypoint = all_waypoints[0]
        curr_x = first_waypoint.x_positions
        curr_y = first_waypoint.y_positions

        color_map = plt.get_cmap("viridis")

        step = 0

        x_tones = np.array(pth.x_tones)
        y_tones = np.array(pth.y_tones)

        x = all_waypoints[0].x_positions
        y = all_waypoints[0].y_positions
        self.clear_paths()
        self.show()

        for action in pth.path:
            if isinstance(action, taskgen.WayPointsAction):
                for start, end in zip(action.way_points[:-1], action.way_points[1:]):
                    x = end.x_positions
                    y = end.y_positions
                    curr_x = start.x_positions
                    curr_y = start.y_positions

                    for (x_tone, x_start, x_end), (y_tone, y_start, y_end) in product(
                        zip(pth.x_tones, curr_x, x), zip(pth.y_tones, curr_y, y)
                    ):
                        self.update_x_bounds(x_start)
                        self.update_x_bounds(x_end)
                        self.update_y_bounds(y_start)
                        self.update_y_bounds(y_end)
                        dx = x_end - x_start
                        dy = y_end - y_start

                        if dx == 0 and dy == 0:
                            continue

                        is_on = (
                            x_tone in self.active_x_tones
                            and y_tone in self.active_y_tones
                        )
                        p = step / (num_arrows - 1) if num_arrows > 1 else 0.0
                        line = self.ax.arrow(
                            x_start,
                            y_start,
                            dx,
                            dy,
                            width=self.arrow_scale,
                            color=color_map(p),
                            length_includes_head=True,
                            linestyle="-" if is_on else (0, (5, 10)),
                            alpha=1.0 if is_on else 0.5,
                            linewidth=1.0 if is_on else 0.5,
                        )

                        line.set_edgecolor(line.get_facecolor())
                        self.curr_path_lines.append(line)

                    if curr_x != x or curr_y != y:
                        step += 1
                        self.show()

            elif isinstance(action, taskgen.TurnOnAction):
                self.active_x_tones.update(x_tones[action.x_tone_indices])
                self.active_y_tones.update(y_tones[action.y_tone_indices])

            elif isinstance(action, taskgen.TurnOffAction):
                self.active_x_tones.difference_update(x_tones[action.x_tone_indices])
                self.active_y_tones.difference_update(y_tones[action.y_tone_indices])

    def set_title(self, title: str) -> None:
        self.ax.set_title(title)

    def show(self) -> None:
        plt.show(block=False)
        self.continue_button.on_clicked(lambda event: setattr(self, "sleeping", False))

        self.sleeping = True
        while self.sleeping:
            plt.pause(self.sleep_time)

    def clear_paths(self) -> None:
        while self.curr_path_lines:
            artist = self.curr_path_lines.pop()
            artist.remove()

        self.ax.set_title("")
        plt.show(block=False)



import bloqade.squin as squin
from bloqade.geometry.dialects import grid

import bloqade.shuttle as qourier


def emit_kernel():

    slm = grid.Grid.from_positions(list(map(float, range(0, 4 * 32, 4))), [20])

    n_gate = int(slm.width // 20) + 1

    gate_slots = grid.Grid.from_positions([0, 3], [0]).repeat(n_gate, 1, 20, 1)
    slm_bounds = slm.x_bounds()
    gate_bounds = gate_slots.x_bounds()
    assert slm_bounds[0] is not None
    assert slm_bounds[1] is not None
    assert gate_bounds[0] is not None
    assert gate_bounds[1] is not None

    slm_mid = (slm_bounds[0] + slm_bounds[1]) / 2
    gate_mid = (gate_bounds[0] + gate_bounds[1]) / 2

    gate_slots = gate_slots.shift(gate_mid - slm_mid, 0)

    @qourier.kernel
    def main():
        qubits = squin.qubit.new(4)
        atoms = qourier.atom.new(slm, qubits)
        ctrls, qargs = qourier.atom.move_next_to(gate_slots, atoms[0::2], atoms[1::2])
        qourier.gate.top_hat_cz(gate_slots)
        results = qourier.atom.measure(ctrls, qubits[::2])
        qourier.atom.reset_position(qargs, qubits[1::2])
        return results

    return main


def test_wrappers():
    ker = emit_kernel()

    assert ker is not None



from kirin import ir, rewrite
from kirin.dialects import func, py

from bloqade.shuttle.dialects import path, schedule
from bloqade.shuttle.rewrite import schedule2path

from .utils import assert_block_equal, sch


def test_rewrite_to_device_call_with_play():
    """Test that the rewrite to device call works correctly."""
    # Create a test value for the device function
    device_fn = ir.TestValue(schedule.DeviceFunctionType)
    input_1 = ir.TestValue()
    input_2 = ir.TestValue()

    # Create a test block with a device function call
    test_block = ir.Block(
        [
            input_1 := py.Constant(1),
            func.Call(
                callee=device_fn,
                inputs=(input_1.result, input_2),
            ),
        ]
    )

    rewrite.Walk(schedule2path.RewriteDeviceCall()).rewrite(test_block)

    expected_block = ir.Block(
        [
            input_1 := py.Constant(1),
            gen_stmt := path.Gen(
                device_fn,
                inputs=(input_1.result, input_2),
            ),
            path.Play(gen_stmt.result),
        ]
    )

    assert_block_equal(test_block, expected_block)


def test_rewrite_to_device_call_without_play():
    """Test that the rewrite to device call works correctly."""
    # Create a test value for the device function
    device_fn = ir.TestValue(schedule.DeviceFunctionType)
    input_1 = ir.TestValue()
    input_2 = ir.TestValue()

    # Create a test block with a device function call
    test_block = ir.Block(
        [
            input_1 := py.Constant(1),
            sch.auto(
                func.Call(
                    callee=device_fn,
                    inputs=(input_1.result, input_2),
                    kwargs=(),
                )
            ),
        ]
    )

    rewrite.Walk(schedule2path.RewriteDeviceCall()).rewrite(test_block)

    expected_block = ir.Block(
        [
            input_1 := py.Constant(1),
            sch.auto(
                path.Gen(
                    device_fn,
                    inputs=(input_1.result, input_2),
                    kwargs=(),
                )
            ),
        ]
    )

    assert_block_equal(test_block, expected_block)



from typing import Any

from bloqade.geometry.dialects import grid
from kirin.lowering import wraps as _wraps

from .stmts import GlobalR, GlobalRz, LocalR, LocalRz, TopHatCZ


@_wraps(TopHatCZ)
def top_hat_cz(
    zone: grid.Grid[Any, Any],
    upper_buffer: float = 3.0,
    lower_buffer: float = 3.0,
) -> None:
    """Apply a top hat CZ gate.

    Args:
        zone (grid.GridType[Any, Any]): The grid zone where the gate is applied.
        upper_buffer (float): The upper buffer distance for the top hat above the zone.
        lower_buffer (float): The lower buffer distance for the top hat below the zone.

    """
    ...


@_wraps(LocalRz)
def local_rz(rotation_angle: float, zone: grid.Grid[Any, Any]) -> None:
    """Apply a local Rz gate to a zone.

    Args:
        rotation_angle (float): The angle of rotation for the Rz gate.
        zone (grid.GridType[Any, Any]): The grid zone where the gate is applied.

    """
    ...


@_wraps(LocalR)
def local_r(
    axis_angle: float, rotation_angle: float, zone: grid.Grid[Any, Any]
) -> None:
    """Apply a local Rz gate to a zone.

    Args:
        rotation_angle (float): The angle of rotation for the Rz gate.
        zone (grid.GridType[Any, Any]): The grid zone where the gate is applied.

    """
    ...


@_wraps(GlobalR)
def global_r(axis_angle: float, rotation_angle: float):
    """Apply a global R gate over all zones.

    Args:
        axis_angle (float): The angle of the axis for the R gate.
        rotation_angle (float): The angle of rotation for the R gate.

    """
    ...


@_wraps(GlobalRz)
def global_rz(rotation_angle: float):
    """Apply a global Rz gate over all zones.

    Args:
        rotation_angle (float): The angle of rotation for the Rz gate.

    """
    ...



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Generic, TypeVar

from typing_extensions import Self

from bloqade.shuttle.arch import ArchSpecInterpreter

if TYPE_CHECKING:
    from bloqade.shuttle.visualizer.renderers import RendererInterface


def default_renderer():
    from bloqade.shuttle.visualizer.renderers.matplotlib import MatplotlibRenderer

    return MatplotlibRenderer()


Plotter = TypeVar("Plotter", bound="RendererInterface")


@dataclass
class PathVisualizer(ArchSpecInterpreter, Generic[Plotter]):
    """Debugging interpreter for visualizing the execution of paths."""

    keys = ["path.visualizer", "spec.interp", "main"]
    renderer: Plotter = field(kw_only=True, default_factory=default_renderer, repr=False)  # type: ignore

    def initialize(self) -> Self:
        for zone_id, zone in self.arch_spec.layout.static_traps.items():
            self.renderer.render_traps(zone, zone_id)

        return super().initialize()



from itertools import repeat
from typing import TypeVar

from bloqade.geometry.dialects import grid
from kirin.dialects import ilist

from bloqade.shuttle import action, schedule, spec
from bloqade.shuttle.prelude import move, tweezer

from .asserts import assert_sorted


def get_spec(
    num_x: int, num_y: int, spacing: float = 10.0, gate_spacing: float = 2.0
) -> spec.ArchSpec:
    """Create a static trap spec with a single zone with pairs of traps oriented
    horizontally.

    Args:
        num_x (int): Number of pairs of traps in the x direction.
        num_y (int): Number of pairs of traps in the y direction.
        spacing (float): Spacing between traps in both directions. Default is 10.0.
        gate_spacing (float): Spacing between gates. Default is 2.0.

    Returns:
        spec.Spec: A specification object containing the layout with a single zone.

    """
    x_spacing = sum(repeat((gate_spacing, spacing), num_x - 1), ()) + (gate_spacing,)
    y_spacing = tuple(repeat(spacing, num_y - 1))

    return spec.ArchSpec(
        layout=spec.Layout(
            {"traps": grid.Grid(x_spacing, y_spacing, 0.0, 0.0)},
            set(["traps"]),
            set(["traps"]),
            set(["traps"]),
        )
    )


NumX = TypeVar("NumX")
NumY = TypeVar("NumY")


@tweezer
def rearrange_impl(
    src_x: ilist.IList[int, NumX],
    src_y: ilist.IList[int, NumY],
    dst_x: ilist.IList[int, NumX],
    dst_y: ilist.IList[int, NumY],
):
    assert len(src_x) == len(
        dst_x
    ), "Source and destination x indices must have the same length."
    assert len(src_y) == len(
        dst_y
    ), "Source and destination y indices must have the same length."

    assert_sorted(src_x)
    assert_sorted(src_y)
    assert_sorted(dst_x)
    assert_sorted(dst_y)

    zone = spec.get_static_trap(zone_id="traps")

    start = grid.sub_grid(zone, src_x, src_y)
    end = grid.sub_grid(zone, dst_x, dst_y)

    def parking_x(index: int):
        x_positions = grid.get_xpos(zone)
        return x_positions[index] + 3.0 * (2 * (index % 2) - 1)

    def parking_y_start(index: int):
        start_y = grid.get_ypos(start)[index]
        end_y = grid.get_ypos(end)[index]
        if start_y <= end_y:
            start_y = start_y + 3.0
        else:
            start_y = start_y - 3.0

        return start_y

    def parking_y_end(index: int):
        start_y = grid.get_ypos(start)[index]
        end_y = grid.get_ypos(end)[index]
        if start_y < end_y:
            end_y = end_y - 3.0
        else:
            end_y = end_y + 3.0

        return end_y

    num_y = len(src_y)

    src_parking = grid.from_positions(
        ilist.map(parking_x, src_x), ilist.map(parking_y_start, ilist.range(num_y))
    )
    dst_parking = grid.from_positions(
        ilist.map(parking_x, dst_x), ilist.map(parking_y_end, ilist.range(num_y))
    )
    mid_pos = grid.from_positions(
        grid.get_xpos(src_parking), grid.get_ypos(dst_parking)
    )

    action.set_loc(start)
    action.turn_on(action.ALL, action.ALL)
    action.move(src_parking)
    action.move(mid_pos)
    action.move(dst_parking)
    action.move(end)
    action.turn_off(action.ALL, action.ALL)


@move
def rearrange(
    src_x: ilist.IList[int, NumX],
    src_y: ilist.IList[int, NumY],
    dst_x: ilist.IList[int, NumX],
    dst_y: ilist.IList[int, NumY],
):
    if len(src_x) < 1 or len(dst_x) < 1:
        return

    x_tones = ilist.range(len(src_x))
    y_tones = ilist.range(len(src_y))

    device_fn = schedule.device_fn(rearrange_impl, x_tones, y_tones)
    device_fn(src_x, src_y, dst_x, dst_y)



from typing import Any, TypeVar, overload

from bloqade.geometry.dialects.grid import Grid
from kirin.dialects import ilist
from kirin.lowering import wraps as _wraps

from .stmts import Move, Set, TurnOff, TurnOn


@overload
def turn_on(x_tones: slice, y_tones: slice) -> None: ...
@overload
def turn_on(x_tones: slice, y_tones: ilist.IList[int, Any] | list[int]) -> None: ...
@overload
def turn_on(x_tones: ilist.IList[int, Any] | list[int], y_tones: slice) -> None: ...
@overload
def turn_on(
    x_tones: ilist.IList[int, Any] | list[int],
    y_tones: ilist.IList[int, Any] | list[int],
) -> None: ...
@_wraps(TurnOn)
def turn_on(x_tones, y_tones):
    """Turn on tones at the given x and y locations."""


@overload
def turn_off(x_tones: slice, y_tones: slice) -> None: ...
@overload
def turn_off(x_tones: slice, y_tones: ilist.IList[int, Any] | list[int]) -> None: ...
@overload
def turn_off(x_tones: ilist.IList[int, Any] | list[int], y_tones: slice) -> None: ...
@overload
def turn_off(
    x_tones: ilist.IList[int, Any] | list[int],
    y_tones: ilist.IList[int, Any] | list[int],
) -> None: ...
@_wraps(TurnOff)
def turn_off(x_tones, y_tones):
    """Turn on tones at the given x and y locations."""


Nx = TypeVar("Nx")
Ny = TypeVar("Ny")


@_wraps(Set)
def set_loc(grid: Grid[Nx, Ny]) -> None:
    """Set the location of the aod using a grid."""


@_wraps(Move)
def move(grid: Grid[Nx, Ny]) -> None:
    """Move the aod from its current location to the given location."""


ALL = slice(None)



from typing import Any, TypeVar

from bloqade.geometry.dialects import grid
from kirin.dialects import ilist
from kirin.lowering import wraps as _wraps

from .stmts import Fill, GetParent, Vacate
from .types import FilledGrid

Nx = TypeVar("Nx")
Ny = TypeVar("Ny")


@_wraps(Vacate)
def vacate(
    zone: grid.Grid[Nx, Ny],
    vacancies: ilist.IList[tuple[int, int], Any],
) -> FilledGrid[Nx, Ny]: ...


@_wraps(Fill)
def fill(
    zone: grid.Grid[Nx, Ny],
    filled: ilist.IList[tuple[int, int], Any],
) -> FilledGrid[Nx, Ny]: ...


@_wraps(GetParent)
def get_parent(filled_grid: FilledGrid[Nx, Ny]) -> grid.Grid[Nx, Ny]: ...


@_wraps(grid.Shift)
def shift(
    grid: FilledGrid[Nx, Ny], x_shift: float, y_shift: float
) -> FilledGrid[Nx, Ny]: ...


@_wraps(grid.Scale)
def scale(
    grid: FilledGrid[Nx, Ny], x_scale: float, y_scale: float
) -> FilledGrid[Nx, Ny]: ...


@_wraps(grid.Repeat)
def repeat(
    grid: FilledGrid[Any, Any],
    x_times: int,
    y_times: int,
    y_spacing: float,
    x_spacing: float,
) -> FilledGrid[Any, Any]: ...



from kirin import interp

from bloqade.shuttle.analysis.runtime import (
    RuntimeAnalysis,
    RuntimeFrame,
)

from ._dialect import dialect
from .stmts import GlobalR, GlobalRz, LocalR, LocalRz, TopHatCZ


@dialect.register(key="runtime")
class HasQuantumRuntimeMethodTable(interp.MethodTable):

    @interp.impl(TopHatCZ)
    @interp.impl(LocalRz)
    @interp.impl(LocalR)
    @interp.impl(GlobalR)
    @interp.impl(GlobalRz)
    def gate(
        self,
        interp: RuntimeAnalysis,
        frame: RuntimeFrame,
        stmt: TopHatCZ | LocalRz | LocalR | GlobalR | GlobalRz,
    ) -> interp.StatementResult[RuntimeFrame]:
        """Handle gate statements and mark the frame as quantum."""
        frame.is_quantum = True
        frame.quantum_stmts.add(stmt)
        return ()



from kirin import interp

from bloqade.shuttle.analysis.runtime import (
    RuntimeAnalysis,
    RuntimeFrame,
)

from ._dialect import dialect
from .stmts import Fill


@dialect.register(key="runtime")
class HasQuantumRuntimeMethodTable(interp.MethodTable):

    @interp.impl(Fill)
    def gate(self, interp: RuntimeAnalysis, frame: RuntimeFrame, stmt: Fill):
        """Handle gate statements and mark the frame as quantum."""
        frame.is_quantum = True
        frame.quantum_stmts.add(stmt)
        return ()



from typing import TypeVar

from kirin import interp

from bloqade.shuttle.dialects.init import Fill, dialect
from bloqade.shuttle.visualizer.interp import PathVisualizer
from bloqade.shuttle.visualizer.renderers import RendererInterface


@dialect.register(key="path.visualizer")
class InitVisualizerMethods(interp.MethodTable):

    Renderer = TypeVar("Renderer", bound=RendererInterface)

    @interp.impl(Fill)
    def fill(self, _interp: PathVisualizer[Renderer], frame: interp.Frame, stmt: Fill):
        return ()



from bloqade.shuttle.prelude import tweezer


@tweezer
def assert_sorted(indices):
    for i in range(1, len(indices)):
        assert indices[i - 1] < indices[i], "Indices must be sorted in ascending order."



from bloqade.geometry.dialects import grid
from kirin import ir
from kirin.dialects import func

from bloqade.shuttle import spec
from bloqade.shuttle.analysis import zone
from bloqade.shuttle.passes.hint_zone import HintZone
from bloqade.shuttle.prelude import move


def return_stmts(kernel: ir.Method) -> list[func.Return]:
    return [
        stmt for stmt in kernel.callable_region.walk() if isinstance(stmt, func.Return)
    ]


def default_spec():
    arch_spec = spec.ArchSpec(
        spec.Layout(
            static_traps={"test": grid.Grid.from_positions([0, 1, 2], [0, 1, 2])},
            fillable=set(["test"]),
            has_cz=set(["test"]),
            has_local=set(["test"]),
        )
    )

    return arch_spec


def test_zone_analysis_unfolded():

    arch_spec = default_spec()

    @move
    def kernel():
        zone = spec.get_static_trap(zone_id="test")
        return zone

    (return_stmt,) = return_stmts(kernel)
    HintZone(move, arch_spec)(kernel)

    assert return_stmt.value.hints["zone.analysis"] == zone.SpecZone("test")


def test_zone_analysis_folded():
    arch_spec = default_spec()

    @move(arch_spec=arch_spec)
    def kernel():
        zone = spec.get_static_trap(zone_id="test")
        return zone

    (return_stmt,) = return_stmts(kernel)
    HintZone(move, arch_spec)(kernel)

    assert return_stmt.value.hints["zone.analysis"] == zone.SpecZone("test")


def test_zone_analysis_invalid():
    arch_spec = default_spec()

    @move(arch_spec=arch_spec)
    def kernel():
        zone = spec.get_static_trap(zone_id="invalid")
        return zone

    (return_stmt,) = return_stmts(kernel)
    HintZone(move, arch_spec)(kernel)

    assert return_stmt.value.hints["zone.analysis"] == zone.InvalidSpecId("invalid")


def test_zone_analysis_subgrid():
    arch_spec = default_spec()

    @move
    def kernel():
        zone = spec.get_static_trap(zone_id="test")
        subgrid = grid.sub_grid(zone, [0, 1], [0, 1])
        return subgrid

    HintZone(move, arch_spec)(kernel)
    (return_stmt,) = return_stmts(kernel)

    assert return_stmt.value.hints["zone.analysis"] == zone.GetSubGridOfZone(
        zone=zone.SpecZone("test"),
        x_indices=zone.Zone.bottom(),
        y_indices=zone.Zone.bottom(),
    )


def test_zone_analysis_subgrid_folded():
    arch_spec = default_spec()

    @move(arch_spec=arch_spec)
    def kernel():
        zone = spec.get_static_trap(zone_id="test")
        subgrid = grid.sub_grid(zone, [0, 1], [0, 1])
        return subgrid

    HintZone(move, arch_spec)(kernel)
    (return_stmt,) = return_stmts(kernel)

    assert return_stmt.value.hints["zone.analysis"] == zone.GetSubGridOfZone(
        zone=zone.SpecZone("test"),
        x_indices=zone.Zone.bottom(),
        y_indices=zone.Zone.bottom(),
    )


def test_zone_analysis_subgrid_invalid():
    arch_spec = default_spec()

    @move
    def kernel():
        zone = spec.get_static_trap(zone_id="invalid")
        subgrid = grid.sub_grid(zone, [0, 1], [0, 1])
        return subgrid

    HintZone(move, arch_spec)(kernel)
    (return_stmt,) = return_stmts(kernel)

    assert return_stmt.value.hints["zone.analysis"] == zone.InvalidZone()


def test_zone_analysis_getitem():
    arch_spec = default_spec()

    @move
    def kernel():
        zone = spec.get_static_trap(zone_id="test")
        return zone[:2, 1:]  # type: ignore

    HintZone(move, arch_spec)(kernel)
    (return_stmt,) = return_stmts(kernel)

    assert return_stmt.value.hints["zone.analysis"] == zone.GetItemOfZone(
        zone=zone.SpecZone("test"),
        index=zone.Zone.bottom(),
    )


def test_zone_analysis_getitem_invalid():
    arch_spec = default_spec()

    @move
    def kernel():
        zone = spec.get_static_trap(zone_id="invalid")
        return zone[:2, 1:]  # type: ignore

    HintZone(move, arch_spec)(kernel)
    (return_stmt,) = return_stmts(kernel)

    assert return_stmt.value.hints["zone.analysis"] == zone.InvalidZone()


def test_zone_analysis_getitem_no_zone():
    arch_spec = default_spec()

    @move
    def kernel():
        a = [1, 2, 3]
        return a[:2]  # type: ignore

    HintZone(move, arch_spec)(kernel)
    (return_stmt,) = return_stmts(kernel)

    assert return_stmt.value.hints["zone.analysis"] == zone.NotZone()



import warnings
from itertools import repeat

from bloqade.geometry.dialects import grid

from bloqade.shuttle import spec

warnings.warn(
    (
        "This module's contents have been moved to `stdlib.layouts.single_col_zone` submodule and"
        " will be removed in 0.8.0"
    ),
    DeprecationWarning,
)


def single_zone_spec(num_x: int, num_y: int, spacing: float = 10.0) -> spec.ArchSpec:
    """Create a static trap spec with a single zone. compatible with the stdlib

    Args:
        num_x (int): Number of traps in the x direction.
        num_y (int): Number of traps in the y direction.
        spacing (float): Spacing between traps in both directions. Default is 10.0.

    Returns:
        spec.Spec: A specification object containing the layout with a single zone.

    """
    x_spacing = tuple(repeat(spacing, num_x - 1))
    y_spacing = tuple(repeat(spacing, num_y - 1))

    return spec.ArchSpec(
        layout=spec.Layout(
            static_traps={"traps": grid.Grid(x_spacing, y_spacing, 0.0, 0.0)},
            fillable=set(["traps"]),
            has_cz=set(["traps"]),
            has_local=set(["traps"]),
        )
    )



from dataclasses import dataclass, field

from kirin import interp, ir
from kirin.analysis import ForwardExtra, ForwardFrame, const
from kirin.dialects import func, scf
from kirin.lattice import EmptyLattice


@dataclass
class RuntimeFrame(ForwardFrame[EmptyLattice]):
    """Frame for quantum runtime analysis.
    This frame is used to track the state of quantum operations within a method.
    """

    quantum_stmts: set[ir.Statement] = field(default_factory=set)
    """Set of quantum statements in the frame."""
    is_quantum: bool = False
    """Whether the frame contains quantum operations."""


class RuntimeAnalysis(ForwardExtra[RuntimeFrame, EmptyLattice]):
    """Forward dataflow analysis to check if a method has quantum runtime.

    This analysis checks if a method contains any quantum runtime operations.
    It is used to determine if the method can be executed on a quantum device.
    """

    keys = ["runtime"]
    lattice = EmptyLattice

    def eval_stmt_fallback(self, frame: RuntimeFrame, stmt: ir.Statement):
        return tuple(self.lattice.top() for _ in stmt.results)

    def initialize_frame(
        self, code: ir.Statement, *, has_parent_access: bool = False
    ) -> RuntimeFrame:
        return RuntimeFrame(code, has_parent_access=has_parent_access)

    def run_method(self, method: ir.Method, args: tuple[EmptyLattice, ...]):
        return self.run_callable(method.code, (self.lattice.bottom(),) + args)

    def has_quantum_runtime(self, method: ir.Method) -> bool:
        """Return True if the method has quantum runtime operations, False otherwise."""
        frame, _ = self.run_analysis(method)
        return frame.is_quantum


@scf.dialect.register(key="runtime")
class Scf(interp.MethodTable):

    @interp.impl(scf.IfElse)
    def ifelse(self, _interp: RuntimeAnalysis, frame: RuntimeFrame, stmt: scf.IfElse):
        # If either branch is quantum, the whole ifelse is quantum
        with _interp.new_frame(stmt, has_parent_access=True) as then_frame:
            then_result = _interp.run_ssacfg_region(
                then_frame, stmt.then_body, (_interp.lattice.top(),)
            )

        with _interp.new_frame(stmt, has_parent_access=True) as else_frame:
            else_result = _interp.run_ssacfg_region(
                else_frame, stmt.else_body, (_interp.lattice.top(),)
            )

        frame.is_quantum = (
            frame.is_quantum or then_frame.is_quantum or else_frame.is_quantum
        )
        frame.quantum_stmts.update(then_frame.quantum_stmts, else_frame.quantum_stmts)
        match (then_result, else_result):
            case (interp.ReturnValue(), tuple()):
                return else_result
            case (tuple(), interp.ReturnValue()):
                return then_result
            case (tuple(), tuple()):
                return tuple(
                    then_result.join(else_result)
                    for then_result, else_result in zip(then_result, else_result)
                )
            case _:
                return tuple(_interp.lattice.top() for _ in stmt.results)

    @interp.impl(scf.For)
    def for_loop(self, _interp: RuntimeAnalysis, frame: RuntimeFrame, stmt: scf.For):
        args = (_interp.lattice.top(),) * (len(stmt.initializers) + 1)
        with _interp.new_frame(stmt, has_parent_access=True) as body_frame:
            result = _interp.run_ssacfg_region(
                body_frame, stmt.body, (_interp.lattice.bottom(),)
            )

        frame.is_quantum = frame.is_quantum or body_frame.is_quantum
        frame.quantum_stmts.update(body_frame.quantum_stmts)
        if isinstance(result, interp.ReturnValue) or result is None:
            return args[1:]
        else:
            return tuple(arg.join(res) for arg, res in zip(args[1:], result))

    @interp.impl(scf.Yield)
    def yield_stmt(
        self, _interp: RuntimeAnalysis, frame: RuntimeFrame, stmt: scf.Yield
    ):
        return interp.YieldValue(frame.get_values(stmt.args))


@func.dialect.register(key="runtime")
class Func(interp.MethodTable):

    @interp.impl(func.Invoke)
    def invoke(self, _interp: RuntimeAnalysis, frame: RuntimeFrame, stmt: func.Invoke):
        args = (_interp.lattice.top(),) * len(stmt.inputs)
        callee_frame, result = _interp.run_method(stmt.callee, args)
        frame.is_quantum = frame.is_quantum or callee_frame.is_quantum
        return (result,)

    @interp.impl(func.Call)
    def call(self, _interp: RuntimeAnalysis, frame: RuntimeFrame, stmt: func.Call):
        # Check if the called method is quantum
        callee_result = stmt.callee.hints.get("const")
        args = (_interp.lattice.top(),) * len(stmt.inputs)
        if (
            isinstance(callee_result, const.PartialLambda)
            and (trait := callee_result.code.get_trait(ir.CallableStmtInterface))
            is not None
        ):
            body = trait.get_callable_region(callee_result.code)
            with _interp.new_frame(stmt) as callee_frame:
                result = _interp.run_ssacfg_region(callee_frame, body, args)
        else:
            raise InterruptedError("Dynamic method calls are not supported")

        frame.is_quantum = frame.is_quantum or callee_frame.is_quantum
        return (result,)

    @interp.impl(func.Return)
    def return_stmt(
        self, _interp: RuntimeAnalysis, frame: RuntimeFrame, stmt: func.Return
    ):
        return interp.ReturnValue(frame.get_values(stmt.results))



from kirin import ir, rewrite

from bloqade.shuttle.dialects import path, schedule
from bloqade.shuttle.rewrite import schedule2path

from .utils import assert_block_equal, pth, sch


def test_rewrite_no_nesting():
    device_fn = ir.TestValue(schedule.DeviceFunctionType)
    input_1 = ir.TestValue()
    input_2 = ir.TestValue()

    test_block = ir.Block(
        [
            sch.parallel(
                path.Gen(device_fn, (input_1, input_2)),
                path.Gen(device_fn, (input_2, input_1)),
            )
        ]
    )

    expected_block = ir.Block(
        [
            path1 := path.Gen(device_fn, (input_1, input_2)),
            path2 := path.Gen(device_fn, (input_2, input_1)),
            ppath := pth.parallel(
                path1.result,
                path2.result,
            ),
            path.Play(ppath.result),
        ]
    )

    rewrite.Walk(schedule2path.RewriteScheduleRegion()).rewrite(test_block)

    assert_block_equal(test_block, expected_block)


def test_rewrite_no_nesting_1_level():
    device_fn = ir.TestValue(schedule.DeviceFunctionType)
    input_1 = ir.TestValue()
    input_2 = ir.TestValue()
    input_3 = ir.TestValue()
    input_4 = ir.TestValue()

    test_block = ir.Block(
        [
            sch.parallel(
                sch.auto(
                    path.Gen(device_fn, (input_1,)),
                    sch.parallel(
                        path.Gen(device_fn, (input_2,)),
                        path.Gen(device_fn, (input_3,)),
                    ),
                ),
                sch.auto(path.Gen(device_fn, (input_4,))),
            )
        ]
    )

    expected_block = ir.Block(
        [
            path1 := path.Gen(device_fn, (input_1,)),
            path2 := path.Gen(device_fn, (input_2,)),
            path3 := path.Gen(device_fn, (input_3,)),
            parallel1 := pth.parallel(path2.result, path3.result),
            auto0 := pth.auto(path1.result, parallel1.result),
            path4 := path.Gen(device_fn, (input_4,)),
            auth1 := pth.auto(path4.result),
            ppath := pth.parallel(auto0.result, auth1.result),
            path.Play(ppath.result),
        ]
    )

    rewrite.Walk(schedule2path.RewriteScheduleRegion()).rewrite(test_block)

    assert_block_equal(test_block, expected_block)


if __name__ == "__main__":
    test_rewrite_no_nesting()
    test_rewrite_no_nesting_1_level()
    print("Test passed!")



from typing import TypeVar

from bloqade.geometry.dialects import grid
from kirin.dialects import ilist

from bloqade.shuttle import action, schedule
from bloqade.shuttle.prelude import move, tweezer

NumX = TypeVar("NumX")
NumY = TypeVar("NumY")
NumWaypoints = TypeVar("NumWaypoints")


@tweezer
def move_by_waypoints_kernel(
    waypoints: ilist.IList[grid.Grid[NumX, NumY], NumWaypoints],
    pick: bool,
    drop: bool,
):
    """Pick up the tweezer at the specified location."""
    action.set_loc(waypoints[0])
    if pick:
        action.turn_on(action.ALL, action.ALL)

    num_waypoints = len(waypoints)
    for i in range(1, num_waypoints):
        action.move(waypoints[i])

    if drop:
        action.turn_off(action.ALL, action.ALL)


@move
def move_by_waypoints(
    waypoints: ilist.IList[grid.Grid[NumX, NumY], NumWaypoints],
    pick: bool = True,
    drop: bool = True,
):
    """Move the tweezer by a list of waypoints.

    Args:
        waypoints (ilist.IList[grid.Grid[NumX, NumY], NumWaypoints]): The waypoints to move to.
        pick (bool): Whether to pick up the tweezer at the first waypoint. Defaults to True.
        drop (bool): Whether to drop the tweezer at the last waypoint. Defaults to True.

    """
    if len(waypoints) < 1:
        return

    shape = grid.shape(waypoints[0])
    device_kernel = schedule.device_fn(
        move_by_waypoints_kernel,
        ilist.range(shape[0]),
        ilist.range(shape[1]),
    )
    device_kernel(waypoints, pick, drop)



from itertools import repeat
from typing import Any, TypeVar

from bloqade.geometry.dialects import grid
from kirin.dialects import ilist

from bloqade.shuttle import action, gate, schedule, spec
from bloqade.shuttle.prelude import move, tweezer

from .asserts import assert_sorted


def get_spec(num_x: int, num_y: int, spacing: float = 10.0) -> spec.ArchSpec:
    """Create a static trap spec with a single zone. compatible with the stdlib

    Args:
        num_x (int): Number of traps in the x direction.
        num_y (int): Number of traps in the y direction.
        spacing (float): Spacing between traps in both directions. Default is 10.0.

    Returns:
        spec.Spec: A specification object containing the layout with a single zone.

    """
    x_spacing = tuple(repeat(spacing, num_x - 1))
    y_spacing = tuple(repeat(spacing, num_y - 1))

    return spec.ArchSpec(
        layout=spec.Layout(
            {"traps": grid.Grid(x_spacing, y_spacing, 0.0, 0.0)},
            set(["traps"]),
            set(["traps"]),
            set(["traps"]),
        )
    )


NumX = TypeVar("NumX")
NumY = TypeVar("NumY")


@tweezer
def single_zone_move_cz(
    zone: grid.Grid[Any, Any],
    ctrl_x_ids: ilist.IList[int, NumX],
    ctrl_y_ids: ilist.IList[int, NumY],
    qarg_x_ids: ilist.IList[int, NumX],
    qarg_y_ids: ilist.IList[int, NumY],
    shift_x: float,
    shift_y: float,
):
    assert len(ctrl_x_ids) == len(
        qarg_x_ids
    ), "Control and target x indices must have the same length."

    assert len(ctrl_y_ids) == len(
        qarg_y_ids
    ), "Control and target y indices must have the same length."

    assert_sorted(ctrl_x_ids)
    assert_sorted(ctrl_y_ids)
    assert_sorted(qarg_x_ids)
    assert_sorted(qarg_y_ids)

    start = grid.sub_grid(zone, ctrl_x_ids, ctrl_y_ids)
    target_atoms = grid.sub_grid(zone, qarg_x_ids, qarg_y_ids)

    end = grid.shift(target_atoms, shift_x, shift_y)
    first_pos = grid.shift(start, shift_x, shift_y)
    second_pos = grid.from_positions(grid.get_xpos(end), grid.get_ypos(first_pos))

    action.set_loc(start)
    action.turn_on(action.ALL, action.ALL)
    action.move(first_pos)
    action.move(second_pos)
    action.move(end)


@move
def default_move_cz_impl(
    zone: grid.Grid[Any, Any],
    x_shift: float,
    y_shift: float,
    ctrl_x_ids: ilist.IList[int, NumX],
    ctrl_y_ids: ilist.IList[int, NumY],
    qarg_x_ids: ilist.IList[int, NumX],
    qarg_y_ids: ilist.IList[int, NumY],
):
    """Move atoms from the start ids and run cz gate with the atoms at the end ids.

    Args:
        zone (grid.Grid[Any, Any]): The grid representing the trap layout (zone) in which atoms are moved.
        x_shift (float): The amount to shift atoms in the x direction during the move.
        y_shift (float): The amount to shift atoms in the y direction during the move.
        ctrl_x_ids (ilist.IList[int, NumX]): The x-indices of the starting positions.
        ctrl_y_ids (ilist.IList[int, NumY]): The y-indices of the starting positions.
        qarg_x_ids (ilist.IList[int, NumX]): The x-indices of the ending positions.
        qarg_y_ids (ilist.IList[int, NumY]): The y-indices of the ending positions.
    """
    if len(ctrl_x_ids) < 1 or len(qarg_x_ids) < 1:
        return

    fwd_kernel = schedule.device_fn(
        single_zone_move_cz,
        ilist.range(len(ctrl_x_ids)),
        ilist.range(len(ctrl_y_ids)),
    )
    bwd_kernel = schedule.reverse(fwd_kernel)

    fwd_kernel(zone, ctrl_x_ids, ctrl_y_ids, qarg_x_ids, qarg_y_ids, x_shift, y_shift)
    gate.top_hat_cz(zone)
    bwd_kernel(zone, qarg_x_ids, qarg_y_ids, ctrl_x_ids, ctrl_y_ids, x_shift, y_shift)


DEFAULT_X_SHIFT = 2.0
DEFAULT_Y_SHIFT = 2.0


@move
def cz_move(
    ctrl_x_ids: ilist.IList[int, NumX],
    ctrl_y_ids: ilist.IList[int, NumY],
    qarg_x_ids: ilist.IList[int, NumX],
    qarg_y_ids: ilist.IList[int, NumY],
):
    zone = spec.get_static_trap(zone_id="traps")
    default_move_cz_impl(
        zone,
        DEFAULT_X_SHIFT,
        DEFAULT_Y_SHIFT,
        ctrl_x_ids,
        ctrl_y_ids,
        qarg_x_ids,
        qarg_y_ids,
    )



import pytest
from bloqade.geometry.dialects import grid
from kirin import interp, ir, prelude
from kirin.dialects import ilist

from bloqade.shuttle.arch import ArchSpec
from bloqade.shuttle.codegen import taskgen
from bloqade.shuttle.dialects import action
from bloqade.shuttle.prelude import tweezer


class TestWaypointsAction:

    def get_waypoint_action(self):
        return taskgen.WayPointsAction()

    def test_init(self):
        waypoint_action = self.get_waypoint_action()
        assert waypoint_action.way_points == []

    def test_add_waypoint(self):
        waypoint_action = self.get_waypoint_action()
        waypoint_action.add_waypoint(pos := grid.Grid.from_positions([1, 2], [3, 4, 5]))
        assert waypoint_action.way_points == [pos]

    def test_inv(self):
        waypoint_action = self.get_waypoint_action()
        waypoint_action.add_waypoint(
            pos1 := grid.Grid.from_positions([1, 2], [3, 4, 5])
        )
        waypoint_action.add_waypoint(
            pos2 := grid.Grid.from_positions([6, 7], [8, 9, 10])
        )
        inverted = waypoint_action.inv()
        assert inverted.way_points == [pos2, pos1]

    def test_repr(self):
        waypoint_action = self.get_waypoint_action()
        waypoint_action.add_waypoint(grid.Grid.from_positions([1, 2], [3, 4, 5]))
        assert (
            repr(waypoint_action) == f"WayPointsAction({waypoint_action.way_points!r})"
        )


class TestTurnOnAction:
    def get_turn_on_action(self):
        return taskgen.TurnOnXYAction(ilist.IList([1, 2]), ilist.IList([3, 4]))

    def test_init(self):
        turn_on_action = self.get_turn_on_action()
        assert turn_on_action.x_tone_indices == ilist.IList([1, 2])
        assert turn_on_action.y_tone_indices == ilist.IList([3, 4])

    def test_inv(self):
        turn_on_action = self.get_turn_on_action()
        inverted = turn_on_action.inv()
        assert isinstance(inverted, taskgen.TurnOffXYAction)
        assert inverted.x_tone_indices == ilist.IList([1, 2])
        assert inverted.y_tone_indices == ilist.IList([3, 4])


class TestTurnOffAction:
    def get_turn_off_action(self):
        return taskgen.TurnOffXYAction(ilist.IList([1, 2]), ilist.IList([3, 4]))

    def test_init(self):
        turn_off_action = self.get_turn_off_action()
        assert turn_off_action.x_tone_indices == ilist.IList([1, 2])
        assert turn_off_action.y_tone_indices == ilist.IList([3, 4])

    def test_inv(self):
        turn_off_action = self.get_turn_off_action()
        inverted = turn_off_action.inv()
        assert isinstance(inverted, taskgen.TurnOnXYAction)
        assert inverted.x_tone_indices == ilist.IList([1, 2])
        assert inverted.y_tone_indices == ilist.IList([3, 4])


class TestActionMethods:

    @property
    def start_pos(self):
        return grid.Grid.from_positions([1, 2], [3, 4])

    def init_interpreter(self):
        interpreter = taskgen.TraceInterpreter(ArchSpec())
        interpreter.initialize()
        return interpreter

    def init_with_current_pos(self):
        interpreter = self.init_interpreter()
        interpreter.curr_pos = self.start_pos
        interpreter.trace = [taskgen.WayPointsAction([self.start_pos])]
        return interpreter

    def test_move_error(self):
        pos_ssa = ir.TestValue()
        pos = grid.Grid.from_positions([4, 5], [3, 4])
        interpreter = self.init_interpreter()

        with pytest.raises(interp.InterpreterError):
            interpreter.run_stmt(action.Move(pos_ssa), (pos,))

    def test_move(self):
        pos_ssa = ir.TestValue()
        pos = grid.Grid.from_positions([4, 5], [3, 4])
        interpreter = self.init_with_current_pos()
        interpreter.run_stmt(action.Move(pos_ssa), (pos,))

        assert interpreter.trace == [
            taskgen.WayPointsAction([self.start_pos, pos]),
        ]

    def test_set(self):
        pos_ssa = ir.TestValue()
        pos = grid.Grid.from_positions([4, 5], [3, 4])
        interpreter = self.init_interpreter()
        interpreter.run_stmt(action.Set(pos_ssa), (pos,))

        assert interpreter.trace == [
            taskgen.WayPointsAction([pos]),
        ]

    def test_turn_on_error(self):
        x_slice_ssa = ir.TestValue()
        y_slice_ssa = ir.TestValue()
        x_tone_indices = ilist.IList([1, 2])
        y_tone_indices = ilist.IList([3, 4])

        interpreter = self.init_interpreter()

        with pytest.raises(interp.InterpreterError):
            interpreter.run_stmt(
                action.TurnOn(x_slice_ssa, y_slice_ssa),
                (x_tone_indices, y_tone_indices),
            )

    def test_turn_off_error(self):
        x_slice_ssa = ir.TestValue()
        y_slice_ssa = ir.TestValue()
        x_tone_indices = ilist.IList([1, 2])
        y_tone_indices = ilist.IList([3, 4])

        interpreter = self.init_interpreter()

        with pytest.raises(interp.InterpreterError):
            interpreter.run_stmt(
                action.TurnOff(x_slice_ssa, y_slice_ssa),
                (x_tone_indices, y_tone_indices),
            )

    @pytest.mark.parametrize(
        ["StmtType", "ActionType", "x_tone_indices", "y_tone_indices"],
        [
            (
                action.TurnOnXYSlice,
                taskgen.TurnOnXYSliceAction,
                slice(1, 2),
                slice(3, 4),
            ),
            (action.TurnOnXSlice, taskgen.TurnOnXSliceAction, slice(1, 2), [1, 2]),
            (
                action.TurnOnYSlice,
                taskgen.TurnOnYSliceAction,
                ilist.IList([1, 2]),
                slice(3, 4),
            ),
            (
                action.TurnOnXY,
                taskgen.TurnOnXYAction,
                ilist.IList([1, 2]),
                ilist.IList([3, 4]),
            ),
            (
                action.TurnOffXYSlice,
                taskgen.TurnOffXYSliceAction,
                slice(1, 2),
                slice(3, 4),
            ),
            (action.TurnOffXSlice, taskgen.TurnOffXSliceAction, slice(1, 2), [1, 2]),
            (
                action.TurnOffYSlice,
                taskgen.TurnOffYSliceAction,
                ilist.IList([1, 2]),
                slice(3, 4),
            ),
            (
                action.TurnOffXY,
                taskgen.TurnOffXYAction,
                ilist.IList([1, 2]),
                ilist.IList([3, 4]),
            ),
        ],
    )
    def test_intensity_actions(
        self, StmtType, ActionType, x_tone_indices, y_tone_indices
    ):
        x_slice_ssa = ir.TestValue()
        y_slice_ssa = ir.TestValue()

        interpreter = self.init_with_current_pos()
        interpreter.run_stmt(
            StmtType(x_slice_ssa, y_slice_ssa),
            (x_tone_indices, y_tone_indices),
        )

        assert interpreter.trace == [
            taskgen.WayPointsAction([self.start_pos]),
            ActionType(x_tone_indices, y_tone_indices),
            taskgen.WayPointsAction([self.start_pos]),
        ]


def test_reverse_path():
    path: list[taskgen.AbstractAction] = [
        taskgen.WayPointsAction([grid.Grid.from_positions([1, 2], [3, 4])]),
        taskgen.TurnOnXYAction(ilist.IList([1, 2]), ilist.IList([3, 4])),
        taskgen.WayPointsAction([grid.Grid.from_positions([5, 6], [7, 8])]),
        taskgen.TurnOffXYAction(ilist.IList([1, 2]), ilist.IList([3, 4])),
    ]

    assert taskgen.reverse_path(path) == [act.inv() for act in reversed(path)]


def test_interpreter_trace():

    @tweezer
    def test_action(self):
        action.set_loc(grid.from_positions([1.0, 2.0], [3.0, 4.0]))
        action.turn_on(action.ALL, action.ALL)
        action.turn_off(action.ALL, action.ALL)

    interpreter = taskgen.TraceInterpreter(ArchSpec())
    assert interpreter.run_trace(test_action, (), {}) == [
        taskgen.WayPointsAction([grid.Grid.from_positions([1, 2], [3, 4])]),
        taskgen.TurnOnXYSliceAction(action.ALL, action.ALL),
        taskgen.WayPointsAction([grid.Grid.from_positions([1, 2], [3, 4])]),
        taskgen.TurnOffXYSliceAction(action.ALL, action.ALL),
        taskgen.WayPointsAction([grid.Grid.from_positions([1, 2], [3, 4])]),
    ]


def test_interpreter_run_trace_error():
    @prelude.basic_no_opt
    def test_bad_method():
        return None

    interpreter = taskgen.TraceInterpreter(ArchSpec())

    with pytest.raises(ValueError):
        interpreter.run_trace(test_bad_method, (), {})



from kirin.analysis.forward import ForwardFrame
from kirin.interp import MethodTable, impl

from bloqade.shuttle.dialects import spec

from .. import analysis, lattice


@spec.dialect.register(key="zone.analysis")
class SpecImpl(MethodTable):
    @impl(spec.GetStaticTrap)
    def get_static_trap(
        self,
        interp: analysis.ZoneAnalysis,
        frame: ForwardFrame[lattice.Zone],
        stmt: spec.GetStaticTrap,
    ):
        if stmt.zone_id in interp.arch_spec.layout.static_traps:
            return (lattice.SpecZone(stmt.zone_id),)

        return (lattice.InvalidSpecId(stmt.zone_id),)



from dataclasses import dataclass
from typing import final

from kirin import ir
from kirin.ir.attrs.abc import LatticeAttributeMeta
from kirin.lattice.abc import BoundedLattice
from kirin.lattice.mixin import SimpleJoinMixin, SimpleMeetMixin
from kirin.print.printer import Printer


@dataclass
class Zone(
    ir.Attribute,
    SimpleJoinMixin["Zone"],
    SimpleMeetMixin["Zone"],
    BoundedLattice["Zone"],
    metaclass=LatticeAttributeMeta,
):

    @classmethod
    def bottom(cls) -> "Zone":
        return NotZone()

    @classmethod
    def top(cls) -> "Zone":
        return UnknownZone()

    def print_impl(self, printer: Printer) -> None:
        printer.print(self.__class__.__name__ + "()")


@final
@dataclass
class NotZone(Zone):

    def is_subseteq(self, other: Zone) -> bool:
        return True


@final
@dataclass
class UnknownZone(Zone):

    def is_subseteq(self, other: Zone) -> bool:
        return isinstance(other, UnknownZone)


@dataclass
class InvalidZone(Zone):
    def is_subseteq(self, other: Zone) -> bool:
        return isinstance(other, InvalidZone)

    def join(self, other: Zone) -> Zone:
        return self


@dataclass
class InvalidSpecId(InvalidZone):
    spec_id: str

    def is_subseteq(self, other: Zone) -> bool:
        return isinstance(other, InvalidSpecId) and (self.spec_id == other.spec_id)

    def join(self, other: Zone) -> Zone:
        if isinstance(other, InvalidSpecId):
            if self.spec_id == other.spec_id:
                return self
            else:
                return InvalidZone()

        return Zone.bottom()


@dataclass
class SpecZone(Zone):
    spec_id: str

    def is_subseteq(self, other: Zone) -> bool:
        return isinstance(other, SpecZone) and (self.spec_id == other.spec_id)


@dataclass
class GetItemLike:
    zone: Zone


@dataclass
class GetItemOfZone(GetItemLike, Zone):
    index: Zone

    def is_subseteq(self, other: Zone) -> bool:
        return (
            isinstance(other, GetItemOfZone)
            and self.zone.is_subseteq(other.zone)
            and self.index.is_subseteq(other.index)
        )


@dataclass
class GetSubGridOfZone(GetItemLike, Zone):
    x_indices: Zone
    y_indices: Zone

    def is_subseteq(self, other: Zone) -> bool:
        return (
            isinstance(other, GetSubGridOfZone)
            and self.zone.is_subseteq(other.zone)
            and self.x_indices.is_subseteq(other.x_indices)
            and self.y_indices.is_subseteq(other.y_indices)
        )



from bloqade.geometry.dialects import grid
from bloqade.squin import op, qubit
from kirin import ir
from kirin.dialects import func, ilist
from kirin.passes import Default, Fold, TypeInfer
from kirin.prelude import structural
from kirin.rewrite import Walk
from kirin.rewrite.chain import Chain

from bloqade.shuttle import spec as spec_module
from bloqade.shuttle.dialects import (
    action,
    atom,
    filled,
    gate,
    init,
    measure,
    path,
    schedule,
    spec,
)
from bloqade.shuttle.passes.inject_spec import InjectSpecsPass
from bloqade.shuttle.passes.schedule2path import ScheduleToPath
from bloqade.shuttle.rewrite.desugar import DesugarTurnOffRewrite, DesugarTurnOnRewrite


@ir.dialect_group(structural.union([spec, grid, filled, atom, gate, op, qubit]))
def kernel(self):
    def run_pass(
        mt: ir.Method,
        *,
        verify: bool = True,
        fold: bool = True,
        aggressive: bool = False,
        typeinfer: bool = True,
        arch_spec: spec_module.ArchSpec | None = None,
    ) -> None:
        if arch_spec is not None:
            InjectSpecsPass(self, arch_spec=arch_spec, fold=False)(mt)

        Default(
            self,
            verify=verify,
            fold=fold,
            aggressive=aggressive,
            typeinfer=typeinfer,
            no_raise=False,
        )(mt)

    return run_pass


# We dont allow [cf, aod, schedule] appear in move function
@ir.dialect_group(structural.union([action, spec, grid, filled]))
def tweezer(self):
    fold_pass = Fold(self)
    typeinfer_pass = TypeInfer(self)
    # TODO: add validation pass after type inference to check
    #       that the number of xtones and ytones match the decorator
    ilist_desugar = ilist.IListDesugar(self)
    action_desugar_pass = Walk(Chain(DesugarTurnOnRewrite(), DesugarTurnOffRewrite()))

    def run_pass(
        mt: ir.Method,
        *,
        fold: bool = True,
        arch_spec: spec_module.ArchSpec | None = None,
    ) -> None:
        if arch_spec is not None:
            InjectSpecsPass(self, arch_spec=arch_spec, fold=False)(mt)

        if isinstance(mt.code, func.Function):
            new_code = action.TweezerFunction(
                sym_name=mt.code.sym_name,
                signature=mt.code.signature,
                body=mt.code.body,
            )
            mt.code = new_code
        else:
            raise ValueError("Method code must be a Function, cannot be lambda/closure")

        ilist_desugar.fixpoint(mt)

        typeinfer_pass(mt)
        action_desugar_pass.rewrite(mt.code)

        if fold:
            fold_pass(mt)

    return run_pass


# no action allow. can have cf, with addtional spec
@ir.dialect_group(
    structural.union([init, schedule, path, grid, filled, spec, gate, op, measure])
)
def move(self):
    schedule_to_path = ScheduleToPath(self)

    def run_pass(
        mt: ir.Method,
        *,
        verify: bool = True,
        fold: bool = True,
        aggressive: bool = False,
        typeinfer: bool = True,
        arch_spec: spec_module.ArchSpec | None = None,
    ) -> None:
        schedule_to_path(mt)

        if arch_spec is not None:
            InjectSpecsPass(self, arch_spec=arch_spec, fold=False)(mt)

        Default(
            self,
            verify=verify,
            fold=fold,
            aggressive=aggressive,
            typeinfer=typeinfer,
            no_raise=False,
        )(mt)

    return run_pass



from .dialects.action import _interface as action
from .dialects.atom import _interface as atom
from .dialects.filled import _interface as filled
from .dialects.gate import _interface as gate
from .dialects.init import _interface as init
from .dialects.measure import _interface as measure
from .dialects.schedule import _interface as schedule
from .dialects.spec import _interface as spec
from .prelude import kernel as kernel, move as move, tweezer as tweezer

__all__ = [
    "action",
    "atom",
    "gate",
    "init",
    "measure",
    "schedule",
    "spec",
    "filled",
]



from kirin import ir, lowering, types
from kirin.decl import info, statement
from kirin.dialects import func, ilist
from kirin.ir.exception import ValidationError
from kirin.lowering.python.traits import FromPythonWithSingleItem

from bloqade.shuttle.dialects.schedule._dialect import dialect
from bloqade.shuttle.dialects.schedule.types import DeviceFunctionType

NxTones = types.TypeVar("NxTones")
NyTones = types.TypeVar("NyTones")


@statement(dialect=dialect)
class NewTweezerTask(ir.Statement):
    name = "tweezer_task"
    traits = frozenset({ir.Pure()})
    move_fn: ir.SSAValue = info.argument(types.MethodType)
    result: ir.ResultValue = info.result(DeviceFunctionType)


@statement(dialect=dialect)
class NewDeviceFunction(ir.Statement):
    name = "device_function"
    traits = frozenset({lowering.FromPythonCall(), ir.Pure()})
    move_fn: ir.SSAValue = info.argument(types.MethodType)
    x_tones: ir.SSAValue = info.argument(ilist.IListType[types.Int, NxTones])
    y_tones: ir.SSAValue = info.argument(ilist.IListType[types.Int, NyTones])
    result: ir.ResultValue = info.result(DeviceFunctionType)


@statement(dialect=dialect)
class Reverse(ir.Statement):
    name = "reverse"
    traits = frozenset({lowering.FromPythonCall(), ir.Pure()})

    device_fn: ir.SSAValue = info.argument(DeviceFunctionType)
    result: ir.ResultValue = info.result(DeviceFunctionType)


@statement
class ExecutableRegion(ir.Statement):
    traits = frozenset({FromPythonWithSingleItem()})
    body: ir.Region = info.region(multi=False)

    def verify_type(self) -> None:
        for stmt in self.body.blocks[0].stmts:
            match stmt:
                case func.Call(
                    callee=ir.SSAValue() as callee
                ) if callee.type.is_subseteq(DeviceFunctionType):
                    continue
                case _:
                    raise ValidationError(
                        self,
                        "All code inside parallel block must be constant or a call of a device task",
                    )


# TODO: implement lowering with captured values
@statement(dialect=dialect)
class Parallel(ExecutableRegion):
    name = "parallel"


# TODO: implement lowering with captured values
@statement(dialect=dialect)
class Auto(ExecutableRegion):
    name = "auto"



import warnings
from typing import Any, TypeVar

from bloqade.geometry.dialects import grid
from kirin.dialects import ilist

from bloqade.shuttle import action, gate, schedule, spec
from bloqade.shuttle.prelude import move, tweezer

warnings.warn(
    (
        "This module has been moved to `stdlib.layouts.single_col_zone` submodule and"
        " will be removed in 0.8.0"
    ),
    DeprecationWarning,
)

NumX = TypeVar("NumX")
NumY = TypeVar("NumY")


@tweezer
def assert_sorted(indices):
    for i in range(1, len(indices)):
        assert indices[i - 1] < indices[i], "Indices must be sorted in ascending order."


@tweezer
def single_zone_move_cz(
    zone: grid.Grid[Any, Any],
    ctrl_x_ids: ilist.IList[int, NumX],
    ctrl_y_ids: ilist.IList[int, NumY],
    qarg_x_ids: ilist.IList[int, NumX],
    qarg_y_ids: ilist.IList[int, NumY],
    shift_x: float,
    shift_y: float,
):
    assert len(ctrl_x_ids) == len(
        qarg_x_ids
    ), "Control and target x indices must have the same length."

    assert len(ctrl_y_ids) == len(
        qarg_y_ids
    ), "Control and target y indices must have the same length."

    assert_sorted(ctrl_x_ids)
    assert_sorted(ctrl_y_ids)
    assert_sorted(qarg_x_ids)
    assert_sorted(qarg_y_ids)

    start = grid.sub_grid(zone, ctrl_x_ids, ctrl_y_ids)
    target_atoms = grid.sub_grid(zone, qarg_x_ids, qarg_y_ids)

    end = grid.shift(target_atoms, shift_x, shift_y)
    first_pos = grid.shift(start, shift_x, shift_y)
    second_pos = grid.from_positions(grid.get_xpos(end), grid.get_ypos(first_pos))

    action.set_loc(start)
    action.turn_on(action.ALL, action.ALL)
    action.move(first_pos)
    action.move(second_pos)
    action.move(end)


@move
def default_move_cz_impl(
    zone: grid.Grid[Any, Any],
    x_shift: float,
    y_shift: float,
    ctrl_x_ids: ilist.IList[int, NumX],
    ctrl_y_ids: ilist.IList[int, NumY],
    qarg_x_ids: ilist.IList[int, NumX],
    qarg_y_ids: ilist.IList[int, NumY],
):
    """Move atoms from the start ids and run cz gate with the atoms at the end ids.

    Args:
        ctrl_x_ids (ilist.IList[int, NumX]): The x-indices of the starting positions.
        ctrl_y_ids (ilist.IList[int, NumY]): The y-indices of the starting positions.
        qarg_x_ids (ilist.IList[int, NumX]): The x-indices of the ending positions.
        qarg_y_ids (ilist.IList[int, NumY]): The y-indices of the ending positions.
    """
    if len(ctrl_x_ids) < 1 or len(qarg_x_ids) < 1:
        return

    fwd_kernel = schedule.device_fn(
        single_zone_move_cz,
        ilist.range(len(ctrl_x_ids)),
        ilist.range(len(ctrl_y_ids)),
    )
    bwd_kernel = schedule.reverse(fwd_kernel)

    fwd_kernel(zone, ctrl_x_ids, ctrl_y_ids, qarg_x_ids, qarg_y_ids, x_shift, y_shift)
    gate.top_hat_cz(zone)
    bwd_kernel(zone, qarg_x_ids, qarg_y_ids, ctrl_x_ids, ctrl_y_ids, x_shift, y_shift)


DEFAULT_X_SHIFT = 2.0
DEFAULT_Y_SHIFT = 2.0


@move
def default_move_cz(
    ctrl_x_ids: ilist.IList[int, NumX],
    ctrl_y_ids: ilist.IList[int, NumY],
    qarg_x_ids: ilist.IList[int, NumX],
    qarg_y_ids: ilist.IList[int, NumY],
):
    zone = spec.get_static_trap(zone_id="traps")
    default_move_cz_impl(
        zone,
        DEFAULT_X_SHIFT,
        DEFAULT_Y_SHIFT,
        ctrl_x_ids,
        ctrl_y_ids,
        qarg_x_ids,
        qarg_y_ids,
    )



from . import runtime as runtime
from ._dialect import dialect as dialect
from ._interface import (
    global_r as global_r,
    global_rz as global_rz,
    local_r as local_r,
    local_rz as local_rz,
    top_hat_cz as top_hat_cz,
)
from .stmts import (
    GlobalR as GlobalR,
    GlobalRz as GlobalRz,
    LocalR as LocalR,
    LocalRz as LocalRz,
    TopHatCZ as TopHatCZ,
)



from bloqade.geometry.dialects.grid import types as zone_types
from kirin import ir, lowering, types
from kirin.decl import info, statement
from kirin.dialects import func, ilist
from kirin.ir import (
    CallableStmtInterface,
    HasSignature,
    IsolatedFromAbove,
    Region,
    Statement,
    SymbolOpInterface,
)
from kirin.print.printer import Printer

from bloqade.shuttle.dialects.action._dialect import dialect


class TweezerFuncOpCallableInterface(CallableStmtInterface["TweezerFunction"]):

    @classmethod
    def get_callable_region(cls, stmt: ir.Statement) -> Region:
        assert isinstance(stmt, TweezerFunction)
        return stmt.body


@statement(dialect=dialect)
class TweezerFunction(Statement):
    name = "move.func"
    traits = frozenset(
        {
            IsolatedFromAbove(),
            SymbolOpInterface(),
            HasSignature(),
            TweezerFuncOpCallableInterface(),
        }
    )
    sym_name: str = info.attribute()
    signature: func.Signature = info.attribute()
    body: Region = info.region(multi=True)

    def print_impl(self, printer: Printer) -> None:
        with printer.rich(style="red"):
            printer.plain_print(self.name + " ")

        with printer.rich(style="cyan"):
            printer.plain_print(self.sym_name)

        self.signature.print_impl(printer)
        printer.plain_print(" ")
        self.body.print_impl(printer)

        with printer.rich(style="black"):
            printer.plain_print(f" // move.func {self.sym_name}")


# this is used as mediate to lower to flair
@statement(dialect=dialect)
class GetXToneId(Statement):
    name = "get.xtoneid"
    traits = frozenset({lowering.FromPythonCall()})
    idx: ir.SSAValue = info.argument(type=types.Int)
    result: ir.ResultValue = info.result(type=types.Int)


# this is used as mediate to lower to flair
@statement(dialect=dialect)
class GetYToneId(Statement):
    name = "get.ytoneid"
    traits = frozenset({lowering.FromPythonCall()})
    idx: ir.SSAValue = info.argument(type=types.Int)
    result: ir.ResultValue = info.result(type=types.Int)


@statement
class IntensityStatement(ir.Statement):
    x_tones: ir.SSAValue = info.argument(
        type=types.Union(
            ilist.IListType[types.Int, types.TypeVar("L")],
            types.Slice[types.Int],
        )
    )
    y_tones: ir.SSAValue = info.argument(
        type=types.Union(
            ilist.IListType[types.Int, types.TypeVar("L")],
            types.Slice[types.Int],
        )
    )


@statement(dialect=dialect)
class TurnOn(IntensityStatement):
    name = "ramp.on"
    traits = frozenset({lowering.FromPythonCall()})


@statement(dialect=dialect)
class TurnOff(IntensityStatement):
    name = "ramp.off"
    traits = frozenset({lowering.FromPythonCall()})


@statement
class IntensityXYSlice(ir.Statement):
    x_tones: ir.SSAValue = info.argument(types.Slice[types.Int])
    y_tones: ir.SSAValue = info.argument(types.Slice[types.Int])


@statement(dialect=dialect)
class TurnOnXYSlice(IntensityXYSlice):
    name = "ramp.on.slice_xy"
    traits = frozenset({})


@statement(dialect=dialect)
class TurnOffXYSlice(IntensityXYSlice):
    name = "ramp.off.slice_xy"
    traits = frozenset({})


@statement
class IntensityXSlice(ir.Statement):
    x_tones: ir.SSAValue = info.argument(types.Slice[types.Int])
    y_tones: ir.SSAValue = info.argument(ilist.IListType[types.Int, types.Any])


@statement(dialect=dialect)
class TurnOnXSlice(IntensityXSlice):
    name = "ramp.on.slice_x"
    traits = frozenset({})


@statement(dialect=dialect)
class TurnOffXSlice(IntensityXSlice):
    name = "ramp.off.slice_x"
    traits = frozenset({})


@statement
class IntensityYSlice(ir.Statement):
    x_tones: ir.SSAValue = info.argument(ilist.IListType[types.Int, types.Any])
    y_tones: ir.SSAValue = info.argument(types.Slice[types.Int])


@statement(dialect=dialect)
class TurnOnYSlice(IntensityYSlice):
    name = "ramp.on.slice_y"
    traits = frozenset({})


@statement(dialect=dialect)
class TurnOffYSlice(IntensityYSlice):
    name = "ramp.off.slice_y"
    traits = frozenset({})


@statement
class Intensity(ir.Statement):
    x_tones: ir.SSAValue = info.argument(ilist.IListType[types.Int, types.Any])
    y_tones: ir.SSAValue = info.argument(ilist.IListType[types.Int, types.Any])


@statement(dialect=dialect)
class TurnOnXY(Intensity):
    name = "ramp.on.xy"
    traits = frozenset({})


@statement(dialect=dialect)
class TurnOffXY(Intensity):
    name = "ramp.off.xy"
    traits = frozenset({})


NxTones = types.TypeVar("NxTones")
NyTones = types.TypeVar("NyTones")


@statement(dialect=dialect)
class Set(Statement):
    # set location
    name = "set"
    traits = frozenset({lowering.FromPythonCall()})
    grid: ir.SSAValue = info.argument(type=zone_types.GridType[NxTones, NyTones])


@statement(dialect=dialect)
class Move(Statement):
    # move to location
    name = "move"
    traits = frozenset({lowering.FromPythonCall()})
    grid: ir.SSAValue = info.argument(type=zone_types.GridType[NxTones, NyTones])



from typing import Any, ContextManager, ParamSpec

from kirin import ir
from kirin.dialects import ilist
from kirin.lowering import wraps as _wraps

from .stmts import (
    Auto,
    NewDeviceFunction,
    Parallel,
    Reverse,
)
from .types import DeviceFunction, ReverseDeviceFunction

Param = ParamSpec("Param")


@_wraps(NewDeviceFunction)
def device_fn(
    move_fn: ir.Method[Param, None],
    x_tones: list[int] | ilist.IList[int, Any],
    y_tones: list[int] | ilist.IList[int, Any],
) -> DeviceFunction[Param]:
    """Create a device function from a move function.

    Args:
        move_fn (Callable): The move function to be wrapped.
        x_tones (list[int]|): The x tones to be used in the device function.
        y_tones: The y tones to be used in the device function.


    """
    ...


@_wraps(Reverse)
def reverse(
    device_fn: DeviceFunction[Param],
) -> ReverseDeviceFunction[Param]:
    """Create a reverse device function from a device function.

    Args:
        device_fn (DeviceFunction): The device function to be reversed.


    """
    ...


@_wraps(Parallel)
def parallel() -> ContextManager: ...


@_wraps(Auto)
def auto() -> ContextManager: ...



from typing import TypeVar

from kirin import interp

from bloqade.shuttle.dialects.gate import (
    GlobalR,
    GlobalRz,
    LocalR,
    LocalRz,
    TopHatCZ,
    dialect,
)
from bloqade.shuttle.visualizer.interp import PathVisualizer
from bloqade.shuttle.visualizer.renderers import RendererInterface


@dialect.register(key="path.visualizer")
class GateVisualizerMethods(interp.MethodTable):

    Renderer = TypeVar("Renderer", bound="RendererInterface")

    @interp.impl(TopHatCZ)
    def fill(
        self, _interp: "PathVisualizer[Renderer]", frame: interp.Frame, stmt: TopHatCZ
    ):
        _interp.renderer.top_hat_cz(
            frame.get(stmt.zone), stmt.upper_buffer, stmt.lower_buffer
        )
        return ()

    @interp.impl(LocalR)
    def local_r(
        self, _interp: "PathVisualizer[Renderer]", frame: interp.Frame, stmt: LocalR
    ):
        _interp.renderer.local_r(frame.get(stmt.zone))
        return ()

    @interp.impl(LocalRz)
    def local_rz(
        self, _interp: "PathVisualizer[Renderer]", frame: interp.Frame, stmt: LocalRz
    ):
        _interp.renderer.local_rz(frame.get(stmt.zone))
        return ()

    @interp.impl(GlobalR)
    def global_r(
        self, _interp: "PathVisualizer[Renderer]", frame: interp.Frame, stmt: GlobalR
    ):
        _interp.renderer.global_r()
        return ()

    @interp.impl(GlobalRz)
    def global_rz(
        self, _interp: "PathVisualizer[Renderer]", frame: interp.Frame, stmt: GlobalRz
    ):
        _interp.renderer.global_rz()
        return ()



from kirin import ir

from bloqade.shuttle.dialects import path, schedule


def simple_region(*stmts: ir.Statement) -> ir.Region:
    """Create a simple block with the given statements."""
    return ir.Region([ir.Block(stmts)])


def call(callee: ir.SSAValue, inputs: tuple[ir.SSAValue, ...]) -> path.Gen:
    """Create a function call with the given callee and inputs."""
    return path.Gen(callee, inputs, kwargs=())


class sch:

    @classmethod
    def auto(cls, *stmts: ir.Statement) -> schedule.Auto:
        """Create an Auto statement with the given statements."""
        return schedule.Auto(body=simple_region(*stmts))

    @classmethod
    def parallel(cls, *stmts: ir.Statement) -> schedule.Parallel:
        """Create a Parallel statement with the given statements."""
        return schedule.Parallel(body=simple_region(*stmts))


class pth:

    @classmethod
    def auto(cls, *paths: ir.SSAValue) -> path.Auto:
        """Create an Auto statement with the given statements."""
        return path.Auto(paths=paths)

    @classmethod
    def parallel(cls, *paths: ir.SSAValue) -> path.Parallel:
        """Create a Parallel statement with the given statements."""
        return path.Parallel(paths=paths)


def assert_block_equal(actual: ir.Block, expected: ir.Block):
    """Assert that two blocks are structurally equal."""
    try:
        assert actual.is_structurally_equal(expected)
    except AssertionError as e:
        actual.print()
        expected.print()
        raise e



from kirin import interp

from bloqade.shuttle.arch import ArchSpecInterpreter

from ._dialect import dialect
from .stmts import GetStaticTrap


@dialect.register(key="spec.interp")
class ArchSpecMethods(interp.MethodTable):
    @interp.impl(GetStaticTrap)
    def get_static_trap(
        self,
        _interp: ArchSpecInterpreter,
        frame: interp.Frame,
        stmt: GetStaticTrap,
    ):
        if (zone := _interp.arch_spec.layout.static_traps.get(stmt.zone_id)) is None:
            raise interp.InterpreterError("Zone not found in layout.")
        return (zone,)



from typing import cast

from bloqade.geometry.dialects import grid
from kirin import decl, ir, lowering, types
from kirin.decl import info
from kirin.dialects import ilist

from ._dialect import dialect
from .types import MeasurementArrayType, MeasurementResultType

NumX = types.TypeVar("NumX")
NumY = types.TypeVar("NumY")

MeasureReturnType = types.Tuple[types.Vararg(MeasurementArrayType[NumX, NumY])]


@decl.statement(dialect=dialect)
class Measure(ir.Statement):
    name = "measure"

    traits = frozenset({lowering.FromPythonCall()})

    grids: tuple[ir.SSAValue, ...] = info.argument(grid.GridType[NumX, NumY])

    def __init__(self, grids: tuple[ir.SSAValue, ...]):
        result_types = []

        for grid_ssa in grids:
            grid_type = grid_ssa.type
            if (grid_type := cast(types.Generic, grid_type)).is_subseteq(grid.GridType):
                NumX, NumY = grid_type.vars
            else:
                NumX, NumY = types.Any, types.Any

            result_types.append(MeasurementArrayType[NumX, NumY])

        super().__init__(
            args=grids,
            result_types=tuple(result_types),
            args_slice={
                "grids": slice(0, len(grids)),
            },
        )


L = types.TypeVar("L")


@decl.statement(dialect=dialect)
class New(ir.Statement):
    name = "new"

    traits = frozenset({lowering.FromPythonCall(), ir.Pure()})

    num_rows: ir.SSAValue = info.argument(
        NumRows := types.TypeVar("NumRows", bound=types.Int)
    )
    num_cols: ir.SSAValue = info.argument(
        NumCols := types.TypeVar("NumCols", bound=types.Int)
    )
    values: ir.SSAValue = info.argument(ilist.IListType[MeasurementResultType, L])
    result: ir.ResultValue = info.result(MeasurementArrayType[NumRows, NumCols])



from bloqade.geometry.dialects import grid

from bloqade.shuttle.arch import ArchSpec
from bloqade.shuttle.codegen.taskgen import (
    TraceInterpreter,
    TurnOffXYSliceAction,
    TurnOnXYSliceAction,
    WayPointsAction,
)
from bloqade.shuttle.dialects import action
from bloqade.shuttle.prelude import tweezer


def test_trace():

    @tweezer
    def move_fn(x: float, y: float):
        start = grid.from_positions([x], [y])
        end = grid.from_positions([x + 1], [y + 1])

        action.set_loc(start)
        action.turn_on(slice(None), slice(None))
        action.move(end)
        action.turn_off(slice(None), slice(None))

    move_fn.print()

    action_list = TraceInterpreter(ArchSpec()).run_trace(move_fn, (1.0, 2.0), {})

    assert isinstance(action_list, list)

    assert isinstance(action_list[0], WayPointsAction)
    assert isinstance(action_list[1], TurnOnXYSliceAction)
    assert isinstance(action_list[2], WayPointsAction)
    assert isinstance(action_list[3], TurnOffXYSliceAction)
    assert isinstance(action_list[4], WayPointsAction)



from bloqade.geometry.dialects import grid

from bloqade.shuttle.dialects import action
from bloqade.shuttle.prelude import tweezer


def test_move():

    @tweezer
    def k1(a: float, b: float, c: float, d: float, e: float):
        start = grid.from_positions([a, b], [c, d, e])
        end = grid.from_positions([a, b + 4], [c, d, e])

        action.turn_on(action.ALL, action.ALL)
        action.set_loc(start)
        action.move(end)
        action.turn_off(action.ALL, action.ALL)

    assert isinstance(k1.code, action.TweezerFunction)



from typing import Any

from bloqade.geometry.dialects.grid.types import Grid
from kirin.dialects import ilist
from kirin.interp import (
    Frame,
    Interpreter,
    MethodTable,
    impl,
)

from . import stmts
from ._dialect import dialect
from .types import FilledGrid


@dialect.register
class FilledGridMethods(MethodTable):

    @impl(stmts.Vacate)
    def vacate(self, interp: Interpreter, frame: Frame, stmt: stmts.Vacate):
        zone = frame.get_casted(stmt.zone, Grid)
        vacancies = frame.get_casted(stmt.vacancies, ilist.IList[tuple[int, int], Any])
        return (FilledGrid.vacate(zone, vacancies),)

    @impl(stmts.Fill)
    def fill(self, interp: Interpreter, frame: Frame, stmt: stmts.Fill):
        zone = frame.get_casted(stmt.zone, Grid)
        filled = frame.get_casted(stmt.filled, ilist.IList[tuple[int, int], Any])
        return (FilledGrid.fill(zone, filled),)

    @impl(stmts.GetParent)
    def get_parent(self, interp: Interpreter, frame: Frame, stmt: stmts.GetParent):
        filled_grid = frame.get_casted(stmt.filled_grid, FilledGrid)
        return (filled_grid.parent,)



from dataclasses import dataclass, field
from typing import Any

from kirin import types
from kirin.dialects import ilist

from bloqade.shuttle.codegen.taskgen import AbstractAction


@dataclass
class Path:
    x_tones: ilist.IList[int, Any]
    y_tones: ilist.IList[int, Any]
    path: list[AbstractAction] = field(default_factory=list, repr=False)

    def __repr__(self) -> str:
        return f"Path({self.x_tones!r}, {self.y_tones!r}, {self.path!r})"

    def __hash__(self):
        return id(self)


PathType = types.PyClass(Path)



from typing import TypeVar

from kirin import interp

from bloqade.shuttle.dialects.measure import (
    Measure,
    dialect,
)
from bloqade.shuttle.visualizer.interp import PathVisualizer
from bloqade.shuttle.visualizer.renderers import RendererInterface


class MeasureResultRuntime:
    pass


@dialect.register(key="path.visualizer")
class MeasureVisualizerMethods(interp.MethodTable):

    Renderer = TypeVar("Renderer", bound="RendererInterface")

    @interp.impl(Measure)
    def fill(
        self, _interp: "PathVisualizer[Renderer]", frame: interp.Frame, stmt: Measure
    ):
        return (MeasureResultRuntime(),)



from dataclasses import dataclass, field
from itertools import chain
from typing import cast

import numpy as np
from bloqade.geometry.dialects.grid import Grid
from kirin.interp import Interpreter


@dataclass
class Layout:
    static_traps: dict[str, Grid]
    """Abstract base class for layout."""

    fillable: set[str]
    """The set of trap names that are fillable by the sorter."""

    has_cz: set[str]
    """The set of trap names that can have a CZ gates applied."""

    has_local: set[str]
    """The set of trap names that can have local single qubit gates applied."""

    special_grid: dict[str, Grid] = field(default_factory=dict, kw_only=True)
    """Set of special grid values that are not static traps, but can be used for specific purposes."""

    def __hash__(self):
        return hash(
            (
                frozenset(self.static_traps.items()),
                frozenset(self.fillable),
                frozenset(self.has_cz),
                frozenset(self.has_local),
                frozenset(self.special_grid.items()),
            )
        )

    def __eq__(self, other):
        if not isinstance(other, Layout):
            return NotImplemented
        return (
            self.static_traps == other.static_traps and self.fillable == other.fillable
        )

    def bounding_box(self) -> tuple[float, float, float, float]:
        """Get the bounding box (xmin, xmax, ymin, ymax) of the layout."""
        xmin = float("inf")
        xmax = float("-inf")
        ymin = float("inf")
        ymax = float("-inf")

        for zone in chain(self.static_traps.values(), self.special_grid.values()):
            if zone.x_init is not None:
                xmin = min(xmin, zone.x_init)
                xmax = max(xmax, zone.x_init + zone.width)
            if zone.y_init is not None:
                ymin = min(ymin, zone.y_init)
                ymax = max(ymax, zone.y_init + zone.height)

        if (
            xmin == float("inf")
            or xmax == float("-inf")
            or ymin == float("inf")
            or ymax == float("-inf")
        ):
            raise ValueError("Layout has incomplete bounding box data.")

        return xmin, xmax, ymin, ymax

    @staticmethod
    def _plot_zone(zone: Grid, ax, name: str, **plot_options):
        from matplotlib.axes import Axes  # type: ignore
        from matplotlib.patches import Rectangle  # type: ignore

        ax = cast(Axes, ax)

        xs, ys = np.meshgrid(tuple(zone.x_positions), tuple(zone.y_positions))

        path_collection = ax.scatter(xs.ravel(), ys.ravel(), **plot_options)

        width = zone.width + 4
        height = zone.height + 4
        xy = (zone.x_positions[0] - 2, zone.y_positions[0] - 2)

        patch = ax.add_patch(
            Rectangle(
                xy=xy,
                width=width,
                height=height,
                edgecolor=path_collection.get_edgecolor(),
                facecolor="none",
            )
        )
        text_mid = (zone.x_init or 0.0) + zone.width / 2

        ax.text(
            x=text_mid, y=zone.y_positions[-1] + 4, s=name, ha="center", va="center"
        )

        return path_collection, patch

    def fig(self, ax=None):
        from matplotlib import pyplot as plt  # type: ignore

        if ax is None:
            _, ax = plt.subplots(1, 1)

        for zone_id, zone in self.static_traps.items():
            self._plot_zone(zone, ax, zone_id)

        return ax, plt.gcf()

    def show(self, ax=None):
        # impoting this so that matplotlib is optional
        import matplotlib.pyplot as plt  # type: ignore

        ax, fig = self.fig(ax)
        plt.show()


def _default_layout():
    zone = Grid.from_positions(
        range(16),
        range(16),
    ).scale(10.0, 10.0)

    return Layout(
        static_traps={"traps": zone},
        fillable={"traps"},
        has_cz={"traps"},
        has_local={"traps"},
    )


@dataclass(frozen=True)
class ArchSpec:
    layout: Layout = field(default_factory=_default_layout)  # type: ignore
    float_constants: dict[str, float] = field(default_factory=dict)
    int_constants: dict[str, int] = field(default_factory=dict)

    def __hash__(self):
        return hash(
            (
                self.layout,
                frozenset(self.float_constants.items()),
                frozenset(self.int_constants.items()),
            )
        )


@dataclass
class ArchSpecMixin:
    """Base class for interpreters that require an architecture specification."""

    arch_spec: ArchSpec


@dataclass
class ArchSpecInterpreter(ArchSpecMixin, Interpreter):
    """Interpreter that requires an architecture specification."""

    pass



from bloqade.geometry.dialects import grid
from kirin import decl, ir, lowering, types
from kirin.decl import info
from kirin.dialects import ilist

from ._dialect import dialect
from .types import FilledGridType

NumVacant = types.TypeVar("NumVacant")
Nx = types.TypeVar("Nx")
Ny = types.TypeVar("Ny")


@decl.statement(dialect=dialect)
class Vacate(ir.Statement):
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})

    zone: ir.SSAValue = info.argument(grid.GridType[Nx, Ny])
    vacancies: ir.SSAValue = info.argument(
        ilist.IListType[types.Tuple[types.Int, types.Int], NumVacant]
    )
    result: ir.ResultValue = info.result(FilledGridType[Nx, Ny])


@decl.statement(dialect=dialect)
class Fill(ir.Statement):
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})

    zone: ir.SSAValue = info.argument(grid.GridType[Nx, Ny])
    filled: ir.SSAValue = info.argument(
        ilist.IListType[types.Tuple[types.Int, types.Int], NumVacant]
    )
    result: ir.ResultValue = info.result(FilledGridType[Nx, Ny])


@decl.statement(dialect=dialect)
class GetParent(ir.Statement):
    name = "get_parent"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})

    filled_grid: ir.SSAValue = info.argument(FilledGridType[Nx, Ny])
    result: ir.ResultValue = info.result(grid.GridType[Nx, Ny])



from bloqade.geometry.dialects import grid

from bloqade.shuttle import spec
from bloqade.shuttle.stdlib.spec import single_zone_spec


def test_single_zone_spec():
    arch_spec = single_zone_spec(3, 3, 10.0)

    expected_spec = spec.ArchSpec(
        layout=spec.Layout(
            static_traps={
                "traps": grid.Grid(
                    x_spacing=(10.0, 10.0),
                    y_spacing=(10.0, 10.0),
                    x_init=0.0,
                    y_init=0.0,
                )
            },
            fillable=set(["traps"]),
            has_cz=set(["traps"]),
            has_local=set(["traps"]),
        )
    )

    assert (
        arch_spec == expected_spec
    ), "The generated spec does not match the expected spec."



from bloqade.shuttle.dialects.schedule._dialect import dialect as dialect
from bloqade.shuttle.dialects.schedule.concrete import (
    ScheduleInterpreter as ScheduleInterpreter,
)
from bloqade.shuttle.dialects.schedule.stmts import (
    Auto as Auto,
    ExecutableRegion as ExecutableRegion,
    NewDeviceFunction as NewDeviceFunction,
    NewTweezerTask as NewTweezerTask,
    Parallel as Parallel,
    Reverse as Reverse,
)
from bloqade.shuttle.dialects.schedule.types import (
    DeviceFunction as DeviceFunction,
    DeviceFunctionType as DeviceFunctionType,
    ReverseDeviceFunction as ReverseDeviceFunction,
)

from ._interface import (
    device_fn as device_fn,
    reverse as reverse,
)



from kirin import ir
from kirin.interp import Frame, Interpreter, InterpreterError, MethodTable, impl

from bloqade.shuttle.dialects.schedule import stmts, types
from bloqade.shuttle.dialects.schedule._dialect import dialect


@dialect.register
class ScheduleInterpreter(MethodTable):

    @impl(stmts.NewDeviceFunction)
    def device_fn(
        self, interp: Interpreter, frame: Frame, stmt: stmts.NewDeviceFunction
    ):
        move_fn: ir.Method = frame.get(stmt.move_fn)
        x_tones = frame.get(stmt.x_tones)
        y_tones = frame.get(stmt.y_tones)
        return (
            types.DeviceFunction(move_fn=move_fn, x_tones=x_tones, y_tones=y_tones),
        )

    @impl(stmts.Reverse)
    def reverse(self, interp: Interpreter, frame: Frame, stmt: stmts.Reverse):
        device_fn = frame.get(stmt.device_fn)
        if isinstance(device_fn, types.DeviceFunction):
            return (types.ReverseDeviceFunction(device_task=device_fn),)
        elif isinstance(device_fn, types.ReverseDeviceFunction):
            return (device_fn.device_task,)
        else:
            raise InterpreterError("Invalid device task type")



import abc
from dataclasses import dataclass, field
from functools import cache
from typing import Any, ClassVar, Dict, Optional

from bloqade.geometry.dialects import grid
from kirin import ir
from kirin.dialects import func, ilist
from kirin.interp import Frame, InterpreterError, MethodTable, impl
from kirin.ir.method import Method
from typing_extensions import Self

from bloqade.shuttle.arch import ArchSpecInterpreter
from bloqade.shuttle.dialects import action


class AbstractAction(abc.ABC):
    @abc.abstractmethod
    def inv(self) -> "AbstractAction": ...


@dataclass
class WayPointsAction(AbstractAction):
    way_points: list[grid.Grid] = field(default_factory=list)

    def add_waypoint(self, pos: grid.Grid):
        self.way_points.append(pos)

    def inv(self):
        return WayPointsAction(list(reversed(self.way_points)))

    def __repr__(self):
        return f"WayPointsAction({self.way_points!r})"


@dataclass(frozen=True)
class TurnOnAction(AbstractAction):
    x_tone_indices: Any
    y_tone_indices: Any


@dataclass(frozen=True)
class TurnOffAction(AbstractAction):
    x_tone_indices: Any
    y_tone_indices: Any


@dataclass(frozen=True)
class TurnOnXYAction(TurnOnAction):
    x_tone_indices: ilist.IList[int, Any]
    y_tone_indices: ilist.IList[int, Any]

    def inv(self):
        return TurnOffXYAction(self.x_tone_indices, self.y_tone_indices)


@dataclass(frozen=True)
class TurnOffXYAction(TurnOffAction):
    x_tone_indices: ilist.IList[int, Any]
    y_tone_indices: ilist.IList[int, Any]

    def inv(self):
        return TurnOnXYAction(self.x_tone_indices, self.y_tone_indices)


@dataclass(frozen=True)
class TurnOnXSliceAction(TurnOnAction):
    x_tone_indices: slice
    y_tone_indices: ilist.IList[int, Any]

    def inv(self):
        return TurnOffXSliceAction(self.x_tone_indices, self.y_tone_indices)


@dataclass(frozen=True)
class TurnOffXSliceAction(TurnOffAction):
    x_tone_indices: slice
    y_tone_indices: ilist.IList[int, Any]

    def inv(self):
        return TurnOnXSliceAction(self.x_tone_indices, self.y_tone_indices)


@dataclass(frozen=True)
class TurnOnYSliceAction(TurnOnAction):
    x_tone_indices: ilist.IList[int, Any]
    y_tone_indices: slice

    def inv(self):
        return TurnOffYSliceAction(self.x_tone_indices, self.y_tone_indices)


@dataclass(frozen=True)
class TurnOffYSliceAction(TurnOffAction):
    x_tone_indices: ilist.IList[int, Any]
    y_tone_indices: slice

    def inv(self):
        return TurnOnYSliceAction(self.x_tone_indices, self.y_tone_indices)


@dataclass(frozen=True)
class TurnOnXYSliceAction(TurnOnAction):
    x_tone_indices: slice
    y_tone_indices: slice

    def inv(self):
        return TurnOffXYSliceAction(self.x_tone_indices, self.y_tone_indices)


@dataclass(frozen=True)
class TurnOffXYSliceAction(TurnOffAction):
    x_tone_indices: slice
    y_tone_indices: slice

    def inv(self):
        return TurnOnXYSliceAction(self.x_tone_indices, self.y_tone_indices)


def reverse_path(path: list[AbstractAction]) -> list[AbstractAction]:
    return [action.inv() for action in reversed(path)]


@cache
def _default_dialect():
    from bloqade.shuttle.prelude import (
        tweezer,  # needs to be here to avoid circular import issues
    )

    return tweezer


@dataclass
class TraceInterpreter(ArchSpecInterpreter):
    keys: ClassVar[list[str]] = ["action.tracer", "spec.interp", "main"]
    trace: list[AbstractAction] = field(init=False, default_factory=list)
    curr_pos: Optional[grid.Grid] = field(init=False, default=None)
    dialects: ir.DialectGroup = field(init=False, default_factory=_default_dialect)

    def initialize(self) -> Self:
        self.curr_pos = None
        self.trace = []
        return super().initialize()

    def run_trace(
        self, mt: Method, args: tuple[Any, ...], kwargs: Dict[str, Any]
    ) -> list[AbstractAction]:

        if not isinstance(mt.code, (action.TweezerFunction, func.Lambda)):
            raise ValueError("Method code must be a MoveFunction or Lambda")

        # TODO: use permute_values to get correct order.
        super().run(mt, args=args, kwargs=kwargs)
        return self.trace.copy()


@action.dialect.register(key="action.tracer")
class ActionTracer(MethodTable):

    intensity_actions = {
        action.TurnOnXY: TurnOnXYAction,
        action.TurnOffXY: TurnOffXYAction,
        action.TurnOnXSlice: TurnOnXSliceAction,
        action.TurnOffXSlice: TurnOffXSliceAction,
        action.TurnOnYSlice: TurnOnYSliceAction,
        action.TurnOffYSlice: TurnOffYSliceAction,
        action.TurnOnXYSlice: TurnOnXYSliceAction,
        action.TurnOffXYSlice: TurnOffXYSliceAction,
    }

    @impl(action.TurnOnXY)
    @impl(action.TurnOffXY)
    @impl(action.TurnOnXSlice)
    @impl(action.TurnOffXSlice)
    @impl(action.TurnOnYSlice)
    @impl(action.TurnOffYSlice)
    @impl(action.TurnOnXYSlice)
    @impl(action.TurnOffXYSlice)
    def construct_intensity_actions(
        self,
        interp: TraceInterpreter,
        frame: Frame,
        stmt: action.IntensityStatement,
    ):
        if interp.curr_pos is None:
            raise InterpreterError(
                "Position of AOD not set before turning on/off tones"
            )

        x_tone_indices = frame.get(stmt.x_tones)
        y_tone_indices = frame.get(stmt.y_tones)

        interp.trace.append(
            self.intensity_actions[type(stmt)](
                x_tone_indices if isinstance(x_tone_indices, slice) else x_tone_indices,
                y_tone_indices if isinstance(y_tone_indices, slice) else y_tone_indices,
            )
        )
        interp.trace.append(WayPointsAction(way_points=[interp.curr_pos]))
        return ()

    @impl(action.Move)
    def move(self, interp: TraceInterpreter, frame: Frame, stmt: action.Move):
        if interp.curr_pos is None:
            raise InterpreterError("Position of AOD not set before moving tones")

        assert isinstance(interp.trace[-1], WayPointsAction)

        interp.trace[-1].add_waypoint(pos := frame.get_typed(stmt.grid, grid.Grid))
        if interp.curr_pos.shape != pos.shape:
            raise InterpreterError(
                f"Position of AOD {interp.curr_pos} and target position {pos} have different shapes"
            )
        interp.curr_pos = pos

        return ()

    @impl(action.Set)
    def set(self, interp: TraceInterpreter, frame: Frame, stmt: action.Set):
        pos = frame.get_typed(stmt.grid, grid.Grid)
        interp.trace.append(WayPointsAction([pos]))

        interp.curr_pos = pos

        return ()



from kirin import ir, types
from kirin.rewrite import abc

from bloqade.shuttle.dialects import action


class DesugarTurnOnRewrite(abc.RewriteRule):

    @staticmethod
    def get_stmt_type(is_x_slice: bool, is_y_slice: bool):
        if is_x_slice and is_y_slice:
            return action.TurnOnXYSlice
        elif is_x_slice:
            return action.TurnOnXSlice
        elif is_y_slice:
            return action.TurnOnYSlice
        else:
            return action.TurnOnXY

    def rewrite_Statement(self, node: ir.Statement) -> abc.RewriteResult:
        if isinstance(node, action.TurnOn):
            is_x_slice = node.x_tones.type.is_subseteq(types.Slice)
            is_y_slice = node.y_tones.type.is_subseteq(types.Slice)
            stmt_type = self.get_stmt_type(is_x_slice, is_y_slice)
            node.replace_by(stmt_type(x_tones=node.x_tones, y_tones=node.y_tones))
            return abc.RewriteResult(has_done_something=True)

        return abc.RewriteResult()


class DesugarTurnOffRewrite(abc.RewriteRule):
    @staticmethod
    def get_stmt_type(is_x_slice: bool, is_y_slice: bool):
        if is_x_slice and is_y_slice:
            return action.TurnOffXYSlice
        elif is_x_slice:
            return action.TurnOffXSlice
        elif is_y_slice:
            return action.TurnOffYSlice
        else:
            return action.TurnOffXY

    def rewrite_Statement(self, node: ir.Statement) -> abc.RewriteResult:
        if isinstance(node, action.TurnOff):
            is_x_slice = node.x_tones.type.is_subseteq(types.Slice)
            is_y_slice = node.y_tones.type.is_subseteq(types.Slice)
            stmt_type = self.get_stmt_type(is_x_slice, is_y_slice)
            node.replace_by(stmt_type(x_tones=node.x_tones, y_tones=node.y_tones))
            return abc.RewriteResult(has_done_something=True)

        return abc.RewriteResult()



from bloqade.geometry.dialects.grid import GridType
from kirin import decl, ir, lowering, types
from kirin.decl import info

from ._dialect import dialect


@decl.statement(dialect=dialect)
class TopHatCZ(ir.Statement):
    name = "apply"

    traits = frozenset({lowering.FromPythonCall()})
    zone: ir.SSAValue = info.argument(type=GridType[types.Any, types.Any])
    upper_buffer: float = info.attribute(default=3.0)
    lower_buffer: float = info.attribute(default=3.0)


@decl.statement(dialect=dialect)
class LocalRz(ir.Statement):
    """
    Apply gate op locally
    """

    traits = frozenset({lowering.FromPythonCall()})
    rotation_angle: ir.SSAValue = info.argument(type=types.Float)
    zone: ir.SSAValue = info.argument(type=GridType[types.Any, types.Any])


@decl.statement(dialect=dialect)
class LocalR(ir.Statement):
    """
    Apply gate op locally
    """

    traits = frozenset({lowering.FromPythonCall()})
    axis_angle: ir.SSAValue = info.argument(type=types.Float)
    rotation_angle: ir.SSAValue = info.argument(type=types.Float)
    zone: ir.SSAValue = info.argument(type=GridType[types.Any, types.Any])


@decl.statement(dialect=dialect)
class GlobalR(ir.Statement):
    """
    Apply gate op globally
    """

    traits = frozenset({lowering.FromPythonCall()})
    axis_angle: ir.SSAValue = info.argument(type=types.Float)
    rotation_angle: ir.SSAValue = info.argument(type=types.Float)


@decl.statement(dialect=dialect)
class GlobalRz(ir.Statement):
    """
    Apply gate op globally
    """

    traits = frozenset({lowering.FromPythonCall()})
    rotation_angle: ir.SSAValue = info.argument(type=types.Float)



from ._dialect import dialect as dialect
from ._interface import (
    fill as fill,
    get_parent as get_parent,
    vacate as vacate,
)
from .concrete import FilledGridMethods as FilledGridMethods
from .stmts import Fill as Fill, GetParent as GetParent, Vacate as Vacate
from .types import FilledGrid as FilledGrid, FilledGridType as FilledGridType



from bloqade.geometry.dialects import grid
from kirin import decl, ir, lowering, types
from kirin.decl import info
from kirin.dialects import ilist

from ._dialect import dialect


@decl.statement(dialect=dialect)
class Fill(ir.Statement):
    name = "fill"

    traits = frozenset({lowering.FromPythonCall()})
    locations: ir.SSAValue = info.argument(
        ilist.IListType[grid.GridType[types.Any, types.Any], types.Any]
    )



from typing import cast

from kirin.analysis import const, forward
from kirin.interp import MethodTable, impl

from bloqade.shuttle.codegen import TraceInterpreter, reverse_path
from bloqade.shuttle.dialects import schedule
from bloqade.shuttle.dialects.path import dialect, stmts, types


@dialect.register(key="constprop")
class ConstProp(MethodTable):

    @impl(stmts.Gen)
    def gen(
        self,
        interp: const.Propagate,
        frame: forward.ForwardFrame[const.Result],
        stmt: stmts.Gen,
    ):
        if stmt.arch_spec is None:
            return (const.Result.top(),)

        device_task_prop = frame.get(stmt.device_task)
        if not isinstance(device_task_prop, const.Value):
            return (const.Result.top(),)

        if isinstance(device_task := device_task_prop.data, schedule.DeviceFunction):
            reverse = False
        elif isinstance(device_task, schedule.ReverseDeviceFunction):
            device_task = device_task.device_task
            reverse = True
        else:
            return (const.Result.top(),)

        inputs_results = frame.get_values(stmt.inputs)

        if not all(isinstance(input_, const.Value) for input_ in inputs_results):
            return (const.Result.top(),)

        kwargs = stmt.kwargs
        args = interp.permute_values(
            device_task.move_fn.arg_names, inputs_results, kwargs
        )

        path = TraceInterpreter(stmt.arch_spec).run_trace(
            device_task.move_fn,
            tuple(
                cast(const.Value, arg).data if isinstance(arg, const.Value) else arg
                for arg in args
            ),
            {},
        )

        if reverse:
            path = reverse_path(path)

        return (
            const.Value(
                types.Path(
                    x_tones=device_task.x_tones,
                    y_tones=device_task.y_tones,
                    path=path,
                )
            ),
        )



from bloqade.geometry.dialects import grid
from kirin.analysis.forward import ForwardFrame
from kirin.interp import MethodTable, impl

from .. import analysis, lattice


@grid.dialect.register(key="zone.analysis")
class GridImpl(MethodTable):
    @impl(grid.GetSubGrid)
    def new_grid(
        self,
        interp: analysis.ZoneAnalysis,
        frame: ForwardFrame[lattice.Zone],
        stmt: grid.GetSubGrid,
    ):

        zone = frame.get(stmt.zone)

        if isinstance(zone, lattice.InvalidZone):
            return (lattice.InvalidZone(),)

        return (
            lattice.GetSubGridOfZone(
                zone=zone,
                x_indices=lattice.Zone.bottom(),
                y_indices=lattice.Zone.bottom(),
            ),
        )



from bloqade.geometry.dialects import grid
from kirin import ir, lowering, types
from kirin.decl import info, statement

from bloqade.shuttle.dialects.spec._dialect import dialect


@statement(dialect=dialect)
class GetStaticTrap(ir.Statement):
    name = "get_static_trap"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    zone_id: str = info.attribute()
    result: ir.ResultValue = info.result(grid.GridType[types.Any, types.Any])


@statement(dialect=dialect)
class GetIntConstant(ir.Statement):
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    constant_id: str = info.attribute()
    result: ir.ResultValue = info.result(types.Int)


@statement(dialect=dialect)
class GetFloatConstant(ir.Statement):
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    constant_id: str = info.attribute()
    result: ir.ResultValue = info.result(types.Float)



import pytest
from bloqade.geometry.dialects.grid import Grid

from bloqade.shuttle.arch import Layout


def test_layout():

    layout = Layout(
        {"test": Grid.from_positions(range(16), range(16))},
        {"test"},
        {"test"},
        {"test"},
    )

    assert hash(layout) == hash(
        (
            frozenset(layout.static_traps.items()),
            frozenset(layout.fillable),
            frozenset(layout.has_cz),
            frozenset(layout.has_local),
            frozenset(layout.special_grid.items()),
        )
    )
    assert layout == Layout(
        {"test": Grid.from_positions(range(16), range(16))},
        {"test"},
        {"test"},
        {"test"},
    )
    assert layout != 1
    assert layout.bounding_box() == (0.0, 15.0, 0.0, 15.0)


def test_bounding_box_no_positions():
    layout = Layout(
        {"test": Grid.from_positions([], [])},
        {"test"},
        {"test"},
        {"test"},
    )

    with pytest.raises(ValueError):
        layout.bounding_box()



from bloqade.geometry.dialects import grid
from kirin.dialects import ilist

from bloqade.shuttle import filled
from bloqade.shuttle.prelude import move


def test_vacate():

    @move
    def test():
        zone = grid.from_positions([0, 1, 2], [0, 1, 2])
        filling = ilist.IList([(0, 0), (1, 1), (2, 2)])
        new_zone = filled.vacate(zone, filling)

        shifted_zone = filled.shift(new_zone, 1, 1)
        scaled_zone = filled.scale(shifted_zone, 2, 2)

        return filled.repeat(scaled_zone, 2, 3, 10, 5)

    parent = grid.Grid.from_positions([0, 1, 2], [0, 1, 2])

    assert test() == filled.FilledGrid.vacate(
        parent,
        frozenset([(0, 0), (1, 1), (2, 2)]),
    ).shift(1, 1).scale(2, 2).repeat(2, 3, 10, 5)


def test_shift():
    @move
    def test():
        zone = grid.from_positions([0, 1, 2], [0, 1, 2])
        filling = ilist.IList([(0, 0), (1, 1), (2, 2)])
        new_zone = filled.vacate(zone, filling)

        shifted_zone = filled.shift(new_zone, 1, 1)
        return shifted_zone

    parent = grid.Grid.from_positions([0, 1, 2], [0, 1, 2])

    assert test() == filled.FilledGrid.vacate(
        parent.shift(1, 1),
        frozenset([(0, 0), (1, 1), (2, 2)]),
    )


def test_scale():
    @move
    def test():
        zone = grid.from_positions([0, 1, 2], [0, 1, 2])
        filling = ilist.IList([(0, 0), (1, 1), (2, 2)])
        new_zone = filled.vacate(zone, filling)
        shifted_zone = filled.scale(new_zone, 1, 1)
        return shifted_zone

    parent = grid.Grid.from_positions([0, 1, 2], [0, 1, 2])

    assert test() == filled.FilledGrid.vacate(
        parent.scale(1, 1),
        [(0, 0), (1, 1), (2, 2)],
    )


def test_positions():
    expected_positions = ilist.IList(((0.3, 0.88), (0.3, 0.99), (0.4, 0.99)))
    assert (
        filled.FilledGrid.vacate(
            grid.Grid.from_positions([0.3, 0.4], [0.88, 0.99]), frozenset([(1, 0)])
        ).positions
        == expected_positions
    )


def test_repeat():
    original = filled.FilledGrid.vacate(
        grid.Grid.from_positions([0.3, 0.4], [0.88, 0.99]), frozenset([(0, 1)])
    )

    tiled = original.repeat(2, 3, 10, 5)
    parent = original.parent.repeat(2, 3, 10, 5)

    expected_vacancies = frozenset(
        [
            (0, 1),
            (0, 3),
            (0, 5),
            (2, 1),
            (2, 3),
            (2, 5),
        ]
    )

    assert tiled.parent == parent
    assert tiled.vacancies == expected_vacancies


def test_fill_filled():

    zone = grid.Grid.from_positions([0, 1, 2], [0, 1, 2])
    filled_1 = filled.FilledGrid.fill(zone, [(0, 0)])
    assert filled_1.vacancies == frozenset(
        [(1, 0), (2, 0), (0, 1), (1, 1), (2, 1), (0, 2), (1, 2), (2, 2)]
    )

    filled_2 = filled.FilledGrid.fill(filled_1, [(1, 1), (1, 2)])
    assert filled_2.vacancies == frozenset(
        [(1, 0), (2, 0), (0, 1), (2, 1), (0, 2), (2, 2)]
    )


def test_fill_vacated():
    zone = grid.Grid.from_positions([0, 1, 2], [0, 1, 2])
    filled_1 = filled.FilledGrid.vacate(zone, [(0, 0), (0, 1), (0, 2)])
    assert filled_1.vacancies == frozenset([(0, 0), (0, 1), (0, 2)])

    filled_2 = filled.FilledGrid.fill(filled_1, [(1, 1), (1, 2)])
    assert filled_1.vacancies == frozenset([(0, 0), (0, 1), (0, 2)])

    filled_3 = filled.FilledGrid.fill(filled_2, [(0, 0)])
    assert filled_3.vacancies == frozenset([(0, 1), (0, 2)])


def test_vacate_filled():
    zone = grid.Grid.from_positions([0, 1, 2], [0, 1, 2])
    filled_1 = filled.FilledGrid.fill(zone, [(0, 0), (0, 1), (0, 2)])
    assert filled_1.vacancies == frozenset(
        [(1, 0), (2, 0), (1, 1), (2, 1), (1, 2), (2, 2)]
    )

    filled_2 = filled.FilledGrid.vacate(filled_1, [(1, 1), (1, 2)])
    assert filled_2.vacancies == frozenset(
        [(1, 0), (2, 0), (1, 1), (2, 1), (1, 2), (2, 2)]
    )

    filled_3 = filled.FilledGrid.vacate(filled_2, [(0, 1)])
    assert filled_3.vacancies == frozenset(
        [(1, 0), (2, 0), (1, 1), (2, 1), (1, 2), (2, 2), (0, 1)]
    )


def test_get_view():
    zone = grid.Grid.from_positions([0, 1, 2], [0, 1, 2])
    filled_1 = filled.FilledGrid.vacate(zone, [(0, 0), (0, 1), (0, 2)])

    view = filled_1.get_view(ilist.IList([0, 2]), ilist.IList([0, 2]))

    assert view.vacancies == frozenset([(0, 0), (0, 1)])



import typing
from dataclasses import dataclass

from kirin import ir, types
from kirin.dialects import ilist

Param = typing.ParamSpec("Param")


class AbstractDeviceTask(typing.Generic[Param]):
    def __call__(self, *args: Param.args, **kwargs: Param.kwargs) -> None:
        raise NotImplementedError("This method should not be called directly.")


@dataclass
class DeviceFunction(AbstractDeviceTask[Param]):
    move_fn: ir.Method[Param, None]
    x_tones: ilist.IList[int, typing.Any]
    y_tones: ilist.IList[int, typing.Any]

    def __hash__(self):
        return id(self)


@dataclass
class ReverseDeviceFunction(AbstractDeviceTask[Param]):
    device_task: DeviceFunction[Param]

    def __hash__(self):
        return id(self)


DeviceFunctionType = types.PyClass(AbstractDeviceTask)



from typing import Any, TypeVar, overload

from bloqade.geometry.dialects import grid
from bloqade.types import Qubit
from kirin.dialects import ilist
from kirin.lowering import wraps as _wraps

from .stmts import Measure, Move, MoveNextTo, New, ResetPosition
from .types import Atom

NumQubits = TypeVar("NumQubits")


@overload
def new(zone: grid.Grid, qubits: list[Qubit]) -> ilist.IList[Atom, Any]: ...
@overload
def new(
    zone: grid.Grid, qubits: ilist.IList[Qubit, NumQubits]
) -> ilist.IList[Atom, NumQubits]: ...
@_wraps(New)
def new(zone, qubits):
    """
    Create new atoms in the specified zone with the given qubits.

    Args:
        zone (grid.Grid): The grid zone where atoms will be created.
        qubits (list[Qubit] | ilist.IList[Qubit, NumQubits]): The qubits associated with the new atoms.

    Returns:
        ilist.IList[Atom, NumQubits | Any]: A list of newly created atoms.
    """
    ...


NumAtoms = TypeVar("NumAtoms")


@overload
def move(zone: grid.Grid, atoms: list[Atom]) -> ilist.IList[Atom, Any]: ...
@overload
def move(
    zone: grid.Grid, atoms: ilist.IList[Atom, NumAtoms]
) -> ilist.IList[Atom, NumAtoms]: ...
@_wraps(Move)
def move(zone, atoms) -> ilist.IList[Atom, Any]:
    """
    Create new atoms in the specified zone with the given qubits. Note that the specific locations of the atoms will be determined by the compiler.


    Args:
        zone (grid.Grid): The grid zone where to move the atoms.
        atoms (list[Atom] | ilist.IList[Atom, NumQubits]): The atoms to be moved.

    Returns:
        ilist.IList[Atom, NumQubits | Any]: A list of newly created atoms.
    """
    ...


@overload
def move_next_to(
    zone: grid.Grid[Any, Any], ctrls: list[Atom], qargs: list[Atom]
) -> tuple[ilist.IList[Atom, NumAtoms], ilist.IList[Atom, NumAtoms]]: ...
@overload
def move_next_to(
    zone: grid.Grid[Any, Any], ctrls: ilist.IList[Atom, NumAtoms], qargs: list[Atom]
) -> tuple[ilist.IList[Atom, NumAtoms], ilist.IList[Atom, NumAtoms]]: ...
@overload
def move_next_to(
    zone: grid.Grid[Any, Any], ctrls: list[Atom], qargs: ilist.IList[Atom, NumAtoms]
) -> tuple[ilist.IList[Atom, NumAtoms], ilist.IList[Atom, NumAtoms]]: ...
@overload
def move_next_to(
    zone: grid.Grid[Any, Any],
    ctrls: ilist.IList[Atom, NumAtoms],
    qargs: ilist.IList[Atom, NumAtoms],
) -> tuple[ilist.IList[Atom, NumAtoms], ilist.IList[Atom, NumAtoms]]: ...
@_wraps(MoveNextTo)
def move_next_to(zone, ctrls, qargs):
    """
    Apply an operation to the specified atoms, optionally moving them in the process if the gate operation requires it.
    After the operation, the atoms potentially change their positions which will be reflected in the returned list.

    Args:
        zone (grid.Grid[Any, Any]): The grid zone where the atoms are located.
        ctrls (list[Atom] | ilist.IList[Atom, NumAtoms]): The control atoms for the operation.
        qargs (list[Atom] | ilist.IList[Atom, NumAtoms]): The target atoms for the operation.

    Returns:
        tuple[ilist.IList[Atom, NumAtoms], ilist.IList[Atom, NumAtoms]]:
            A tuple containing two lists of atoms: the control atoms and the target atoms after the operation.
            the first list contains the updated ctrls, and the second list contains the updated qargs.
    """
    ...


@overload
def reset_position(atoms: list[Atom], qubits: list[Qubit]) -> None: ...
@overload
def reset_position(atoms: list[Atom], qubits: ilist.IList[Qubit, NumAtoms]) -> None: ...
@overload
def reset_position(atoms: ilist.IList[Atom, NumAtoms], qubits: list[Qubit]) -> None: ...
@overload
def reset_position(
    atoms: ilist.IList[Atom, NumAtoms], qubits: ilist.IList[Qubit, NumAtoms]
) -> None: ...
@_wraps(ResetPosition)
def reset_position(atoms, qubits) -> None:
    """
    Reset the position of atoms to their initial state, terminating the moves.

    Args:
        atoms (list[Atom] | ilist.IList[Atom, NumAtoms]): The atoms whose positions will be reset.
        qubits (list[Qubit] | ilist.IList[Qubit, NumAtoms]): The qubits associated with the atoms being reset.
    """
    ...


@overload
def measure(atoms: list[Atom], qubits: list[Qubit]) -> ilist.IList[int, Any]: ...
@overload
def measure(
    atoms: ilist.IList[Atom, NumAtoms], qubits: list[Qubit]
) -> ilist.IList[int, NumAtoms]: ...
@overload
def measure(
    atoms: list[Atom], qubits: ilist.IList[Qubit, NumAtoms]
) -> ilist.IList[int, NumAtoms]: ...
@overload
def measure(
    atoms: ilist.IList[Atom, NumAtoms], qubits: ilist.IList[Qubit, NumAtoms]
) -> ilist.IList[int, NumAtoms]: ...
@_wraps(Measure)
def measure(atoms, qubits) -> ilist.IList[int, Any]:
    """
    Perform a destructive measurement on the specified atoms.

    Args:
        atoms (list[Atom] | ilist.IList[Atom, NumAtoms]): The atoms to be measured.
        qubits (list[Qubit] | ilist.IList[Qubit, NumAtoms]): The qubits associated with the atoms being measured.

    Returns:
        ilist.IList[int, NumAtoms | Any]: A list of measurement results, where each result corresponds to an atom.
    """
    ...



from kirin import ir
from kirin.passes import Pass
from kirin.rewrite import (
    Chain,
    CommonSubexpressionElimination,
    DeadCodeElimination,
    Fixpoint,
    Walk,
)

from bloqade.shuttle.rewrite.schedule2path import (
    Canonicalize,
    RewriteAutoInvoke,
    RewriteDeviceCall,
    RewriteScheduleRegion,
)


class ScheduleToPath(Pass):
    """Pass to convert schedule dialect to path dialect."""

    def unsafe_run(self, mt: ir.Method):
        result = Fixpoint(Walk(Canonicalize())).rewrite(mt.code)
        result = (
            Walk(Chain(RewriteAutoInvoke(), RewriteDeviceCall()))
            .rewrite(mt.code)
            .join(result)
        )
        result = Walk(RewriteScheduleRegion()).rewrite(mt.code).join(result)
        result = (
            Fixpoint(
                Walk(Chain(CommonSubexpressionElimination(), DeadCodeElimination()))
            )
            .rewrite(mt.code)
            .join(result)
        )

        return result



from abc import ABC, abstractmethod

from bloqade.geometry.dialects import grid

from bloqade.shuttle.dialects import path


class RendererInterface(ABC):

    def local_r(self, location: grid.Grid) -> None:
        """Render the local render of the given location.

        Args:
            location (grid.Grid): The location to render.

        """

    def local_rz(self, location: grid.Grid) -> None:
        """Render the local render of the given location.

        Args:
            location (grid.Grid): The location to render.

        """

    def global_r(self) -> None:
        """Render the global render of the given location.

        Args:
            location (grid.Grid): The location to render.

        """

    def global_rz(self) -> None:
        """Render the global render of the given location.

        Args:
            location (grid.Grid): The location to render.

        """

    def top_hat_cz(
        self, location: grid.Grid, upper_buffer: float, lower_buffer: float
    ) -> None:
        """Render the top hat CZ of the given location.

        Args:
            location (grid.Grid): The location to render.

        """

    @abstractmethod
    def render_traps(self, traps: grid.Grid, zone_id: str) -> None:
        """Render the given traps.

        Args:
            traps (ilist.IList[grid.Grid, Any]): The traps to render.

        """
        ...

    @abstractmethod
    def set_title(self, title: str) -> None:
        """Set the title of the renderer.

        Args:
            title (str): The title to set.

        """
        ...

    @abstractmethod
    def render_path(self, pth: path.Path) -> None:
        """Render the given path.

        Args:
            pth (path.Path): The path to render.

        """
        ...

    @abstractmethod
    def show(self) -> None:
        """Show all rendered entities."""
        ...

    @abstractmethod
    def clear_paths(self) -> None:
        """Clear the current renderer."""
        ...



from typing import Optional

from kirin import decl, ir
from kirin.decl import info

from bloqade.shuttle.arch import ArchSpec
from bloqade.shuttle.dialects import schedule

from ._dialect import dialect
from .types import PathType


@decl.statement(dialect=dialect)
class Gen(ir.Statement):
    name = "gen"

    traits = frozenset({ir.Pure()})
    # not a fixed type here so just any
    device_task: ir.SSAValue = info.argument(schedule.DeviceFunctionType)
    arch_spec: Optional[ArchSpec] = info.attribute(default=None)
    inputs: tuple[ir.SSAValue, ...] = info.argument()
    kwargs: tuple[str, ...] = info.attribute(default_factory=lambda: ())
    result: ir.ResultValue = info.result(PathType)


@decl.statement(dialect=dialect)
class Parallel(ir.Statement):
    name = "parallel"

    traits = frozenset({ir.Pure()})
    paths: tuple[ir.SSAValue, ...] = info.argument()
    result: ir.ResultValue = info.result(PathType)

    def __init__(self, paths: tuple[ir.SSAValue, ...]):
        super().__init__(
            args=paths,
            result_types=[PathType],
            args_slice={"paths": slice(0, len(paths), 1)},
        )


@decl.statement(dialect=dialect)
class Auto(ir.Statement):
    name = "auto"

    traits = frozenset({ir.Pure()})
    paths: tuple[ir.SSAValue, ...] = info.argument()
    result: ir.ResultValue = info.result(PathType)

    def __init__(self, paths: tuple[ir.SSAValue, ...]):
        super().__init__(
            args=paths,
            result_types=[PathType],
            args_slice={"paths": slice(0, len(paths), 1)},
        )


@decl.statement(dialect=dialect)
class Play(ir.Statement):
    name = "play"

    traits = frozenset({})
    path: ir.SSAValue = info.argument(PathType)



from bloqade.shuttle.dialects.spec._interface import (
    get_float_constant as get_float_constant,
    get_int_constant as get_int_constant,
    get_static_trap as get_static_trap,
)

from ._dialect import dialect as dialect
from .concrete import ArchSpecMethods as ArchSpecMethods
from .stmts import (
    GetFloatConstant as GetFloatConstant,
    GetIntConstant as GetIntConstant,
    GetStaticTrap as GetStaticTrap,
)



from typing import Any

from bloqade.geometry.dialects import grid
from kirin.lowering import wraps as _wraps

from bloqade.shuttle.arch import ArchSpec as ArchSpec, Layout as Layout

from .stmts import GetFloatConstant, GetIntConstant, GetStaticTrap


@_wraps(GetStaticTrap)
def get_static_trap(*, zone_id: str) -> grid.Grid[Any, Any]:
    """Get a static trap by its zone ID."""
    ...


@_wraps(GetIntConstant)
def get_int_constant(*, constant_id: str) -> int: ...


@_wraps(GetFloatConstant)
def get_float_constant(*, constant_id: str) -> float: ...



from dataclasses import dataclass, field

from kirin import ir
from kirin.dialects import ilist, scf
from kirin.ir.method import Method
from kirin.passes import HintConst, Pass, TypeInfer
from kirin.rewrite import (
    Call2Invoke,
    CFGCompactify,
    Chain,
    ConstantFold,
    DeadCodeElimination,
    Fixpoint,
    Inline,
    InlineGetField,
    InlineGetItem,
    Walk,
)
from kirin.rewrite.abc import RewriteResult
from kirin.rewrite.cse import CommonSubexpressionElimination


@dataclass
class AggressiveUnroll(Pass):
    """Fold pass to fold control flow"""

    constprop: HintConst = field(init=False)
    typeinfer: TypeInfer = field(init=False)

    def __post_init__(self):
        self.const_hint = HintConst(self.dialects, no_raise=self.no_raise)
        self.typeinfer = TypeInfer(self.dialects, no_raise=self.no_raise)

    def unsafe_run(self, mt: Method) -> RewriteResult:
        result = RewriteResult()
        result = self.const_hint.unsafe_run(mt).join(result)
        rule = Chain(
            ConstantFold(),
            Call2Invoke(),
            InlineGetField(),
            InlineGetItem(),
            ilist.rewrite.InlineGetItem(),
            DeadCodeElimination(),
            CommonSubexpressionElimination(),
        )
        result = Fixpoint(Walk(rule)).rewrite(mt.code).join(result)
        result = (
            Walk(
                Chain(
                    scf.unroll.PickIfElse(),
                    scf.unroll.ForLoop(),
                    scf.trim.UnusedYield(),
                )
            )
            .rewrite(mt.code)
            .join(result)
        )

        self.typeinfer.unsafe_run(mt)
        result = (
            Walk(Chain(ilist.rewrite.ConstList2IList(), ilist.rewrite.Unroll()))
            .rewrite(mt.code)
            .join(result)
        )

        result = Walk(Inline(self.inline_heuristic)).rewrite(mt.code).join(result)
        result = Walk(Fixpoint(CFGCompactify())).rewrite(mt.code).join(result)
        return result

    @classmethod
    def inline_heuristic(cls, node: ir.Statement) -> bool:
        """The heuristic to decide whether to inline a function call or not.
        inside loops and if-else, only inline simple functions, i.e.
        functions with a single block
        """
        if not isinstance(node.parent_stmt, (scf.For, scf.IfElse)):
            return True  # always inline calls outside of loops and if-else

        if (trait := node.get_trait(ir.CallableStmtInterface)) is None:
            return False  # not a callable, don't inline to be safe
        region = trait.get_callable_region(node)
        return len(region.blocks) == 1



from kirin import ir, rewrite
from kirin.dialects import func

from bloqade.shuttle.rewrite.schedule2path import Canonicalize

from .utils import assert_block_equal, sch


def test_canonicalize_do_nothing():

    callee_1 = ir.TestValue()
    callee_2 = ir.TestValue()
    input_1 = ir.TestValue()
    input_2 = ir.TestValue()

    test_block = ir.Block(
        [
            sch.auto(
                sch.parallel(
                    func.Call(callee_1, (input_1,)),
                    sch.auto(func.Call(callee_1, (input_1,))),
                ),
                sch.parallel(func.Call(callee_2, (input_2,))),
            )
        ]
    )

    expected_block = ir.Block(
        [
            sch.auto(
                sch.parallel(
                    func.Call(callee_1, (input_1,)),
                    sch.auto(func.Call(callee_1, (input_1,))),
                ),
                sch.parallel(func.Call(callee_2, (input_2,))),
            )
        ]
    )

    rewrite.Walk(Canonicalize()).rewrite(test_block)

    assert_block_equal(test_block, expected_block)


def test_canonicalize_flatten_auto():

    callee_1 = ir.TestValue()
    callee_2 = ir.TestValue()
    input_1 = ir.TestValue()
    input_2 = ir.TestValue()

    test_block = ir.Block(
        [
            sch.auto(
                sch.parallel(
                    func.Call(callee_1, (input_1,)),
                    sch.parallel(func.Call(callee_1, (input_1,))),
                ),
                sch.auto(
                    func.Call(callee_2, (input_2,)), func.Call(callee_1, (input_1,))
                ),
            )
        ]
    )

    expected_block = ir.Block(
        [
            sch.auto(
                sch.parallel(
                    func.Call(callee_1, (input_1,)), func.Call(callee_1, (input_1,))
                ),
                func.Call(callee_2, (input_2,)),
                func.Call(callee_1, (input_1,)),
            )
        ]
    )

    rewrite.Walk(Canonicalize()).rewrite(test_block)

    assert_block_equal(test_block, expected_block)



from .analysis import ZoneAnalysis as ZoneAnalysis
from .impl.grid import GridImpl as GridImpl
from .impl.py import ConstantImpl as ConstantImpl, GetIndexImpl as GetIndexImpl
from .impl.spec import SpecImpl as SpecImpl
from .lattice import (
    GetItemOfZone as GetItemOfZone,
    GetSubGridOfZone as GetSubGridOfZone,
    InvalidSpecId as InvalidSpecId,
    InvalidZone as InvalidZone,
    NotZone as NotZone,
    SpecZone as SpecZone,
    UnknownZone as UnknownZone,
    Zone as Zone,
)



import math
from typing import Any, Literal, TypeVar

from bloqade.geometry.dialects import grid
from kirin.dialects import ilist

from bloqade.shuttle import action, gate, init, measure, schedule
from bloqade.shuttle.passes.fold import AggressiveUnroll
from bloqade.shuttle.prelude import move, tweezer

NMove = TypeVar("NMove")


@tweezer
def entangle_move(
    mem_zone: grid.Grid[Any, Literal[1]],
    gate_zone: grid.Grid[NMove, Literal[2]],
    ctrl_ids: ilist.IList[int, NMove],
    qarg_ids: ilist.IList[int, NMove],
    gate_ids: ilist.IList[int, NMove],
):

    mem_y = grid.get_ypos(mem_zone)[0]
    ctrl_start = grid.get_xpos(grid.sub_grid(mem_zone, ctrl_ids, [0]))
    qarg_start = grid.get_xpos(grid.sub_grid(mem_zone, qarg_ids, [0]))

    pos_1 = grid.from_positions(ctrl_start, [mem_y, mem_y])
    pos_2 = grid.shift(grid.from_positions(ctrl_start, [mem_y - 4.0, mem_y]), 2.0, 0.0)
    pos_3 = grid.from_positions(qarg_start, [mem_y - 4.0, mem_y])
    gate_pos = grid.sub_grid(gate_zone, gate_ids, [0, 1])

    action.set_loc(pos_1)
    action.turn_on(action.ALL, [0])
    action.move(pos_2)
    action.move(pos_3)
    action.turn_on([], [1])
    action.move(gate_pos)


@move
def apply_h(zone: grid.Grid[Any, Any]):
    # TODO: This is wrong and needs to be fixed.
    gate.global_r(math.pi / 2.0, -math.pi / 2.0)
    gate.local_rz(math.pi / 2.0, zone)
    gate.global_r(0.0, -math.pi / 2.0)


@move
def apply_cx(gate_zone: grid.Grid[Any, Any]):
    # TODO: This is wrong and needs to be fixed.
    gate.global_r(0.0, math.pi)
    gate.top_hat_cz(gate_zone)
    gate.global_r(0.0, -math.pi)


@move
def run_entangle_move(
    mem_zone: grid.Grid[Any, Literal[1]],
    gate_zone: grid.Grid[Any, Literal[2]],
    ctrl_ids: ilist.IList[int, NMove],
    qarg_ids: ilist.IList[int, NMove],
    gate_ids: ilist.IList[int, NMove],
):

    num = len(ctrl_ids)
    xtones = ilist.range(num)
    ytones = [0, 1]

    dtask = schedule.device_fn(entangle_move, xtones, ytones)
    rev_dtask = schedule.reverse(dtask)

    dtask(mem_zone, gate_zone, ctrl_ids, qarg_ids, gate_ids)
    apply_cx(gate_zone)
    rev_dtask(mem_zone, gate_zone, ctrl_ids, qarg_ids, gate_ids)


N = TypeVar("N")


@move
def ghz_prep_steps(
    qubit_ids: ilist.IList[int, N],
    gate: grid.Grid[Any, Literal[2]],
    mem: grid.Grid[Any, Literal[1]],
):

    n_qubits = len(qubit_ids)
    gate_shape = grid.shape(gate)
    gate_width = gate_shape[0]

    jobs = []
    # calculate the qubits in each layer then split up into groups based on the size of the gate
    for depth in range(n_qubits):
        step = n_qubits // (2**depth)
        half_step = step // 2
        if half_step > 0:
            ctrl_qubits = qubit_ids[:-half_step:half_step]
            qarg_qubits = qubit_ids[half_step::half_step]

            num_gates = len(ctrl_qubits)
            num_substeps = num_gates // gate_width + 1

            for i in range(num_substeps):
                start = i * gate_width
                stop = start + gate_width

                if stop > num_gates:
                    stop = num_gates

                if stop > start:
                    ctrl_ids = ctrl_qubits[start:stop]
                    qarg_ids = qarg_qubits[start:stop]
                    gate_ids = ilist.range(stop - start)
                    jobs = jobs + [(ctrl_ids, qarg_ids, gate_ids)]

    for i in range(len(jobs)):
        ctrl_ids = jobs[i][0]
        qarg_ids = jobs[i][1]
        gate_ids = jobs[i][2]
        run_entangle_move(mem, gate, ctrl_ids, qarg_ids, gate_ids)


def run_ghz():
    spacing = 4
    num_sites = 32
    mem = grid.Grid.from_positions(
        list(map(float, range(0, spacing * num_sites, spacing))), [20.0]
    )

    num_gate = int(mem.width // 20) + 1
    gate = grid.Grid.from_positions(
        list(map(float, range(0, num_gate * 20, 20))), [0.0, 3.0]
    )

    mem_bounds = mem.x_bounds()
    assert mem_bounds[0] is not None
    assert mem_bounds[1] is not None

    mid_mem = (mem.x_positions[0] + mem.x_positions[-1]) / 2
    mid_gate = (gate.x_positions[0] + gate.x_positions[-1]) / 2

    gate_shift = mid_mem - mid_gate

    gate = gate.shift(gate_shift, 0.0)

    @move
    def main():
        init.fill([mem])
        ghz_prep_steps([1, 2, 4, 5, 7, 9, 14, 28, 29, 32], gate, mem)  # type: ignore
        return measure.measure((mem,))

    return main


def test_fold():
    main = run_ghz()
    AggressiveUnroll(main.dialects).fixpoint(main)
    main.print()


if __name__ == "__main__":
    test_fold()
